(** Definition and properties of the mathematical O.
    f ={x -> 0} O(g) becomes O_P(f,g) with P being conditions on acceptable x *)

Require Export R_two.

Open Local Scope nat_scope.
Open Local Scope R_scope.

Set Implicit Arguments.

Section BigOuP.

(** Here is the definition of the O, 
    where dX:R*R -> 0
    and which is uniform for X in A (may be an interval, a box,...) *) 
Variable A:Type.

Definition OuP  (P: R*R-> Prop) (f: A -> R*R  ->  R)  (g: R * R -> R)  :=
  { alp : R & { C : R | 
  0 < alp /\ 0 < C /\
  forall X:A, forall dX : R * R, 
  norm_l2 dX < alp -> P dX ->
  Rabs (f X dX) <= C * Rabs (g dX)}}.   
    

(** We then have a vector space	*)

Lemma OuP_is_EV_zero : 
  forall P, forall g : R * R -> R,
  OuP P (fun (X:A) dX => 0) g.
Proof.
unfold OuP; intros g; exists 1; exists 1; repeat split; auto with real.
intros; rewrite Rabs_R0.
apply Rmult_le_pos; auto with real.
apply Rabs_pos.
Qed.

Lemma OuP_is_EV_plus :
  forall P f1 f2 g,
  OuP P f1 g -> OuP P f2 g ->
  OuP P (fun a => add2_fun (f1 a) (f2 a)) g.  
Proof.
unfold OuP;
 intros P f1 f2 g (alp1, (C1, (H11, (H12, H1)))) (alp2, (C2, (H21, (H22, H2)))).
split with (Rmin alp1 alp2); split with (C1 + C2); repeat split.
unfold Rmin; case (Rle_dec alp1 alp2); auto.
apply Rplus_lt_pos; auto.
intros a dX H H'; unfold add2_fun;
  apply Rle_trans with (Rabs (f1 a dX) + Rabs (f2 a dX)).
apply Rabs_triang.
rewrite Rmult_plus_distr_r; apply Rplus_le_compat.
apply H1; auto.
apply Rlt_le_trans with (1 := H).
apply Rmin_l.
apply H2; auto.
apply Rlt_le_trans with (1 := H).
apply Rmin_r.
Qed.


Lemma OuP_is_EV_mult :
  forall P (f:A -> R*R -> R) g l,
  OuP P f g ->
  OuP P (fun a:A => scalar_multiplyA_fun _ l (f a)) g.
Proof. 
unfold OuP, scalar_multiplyA_fun; intros P f g l (alp, (C, (H1, (H2, H3)))).
case (Req_EM_T l 0); intros H4.
exists alp; exists C.
repeat split; auto; intros;
  rewrite H4;
  ring_simplify (0 * f X dX);
  rewrite Rabs_R0;
  apply Rle_trans with (Rabs (f X dX)).
apply Rabs_pos.
apply H3; auto.
exists alp; exists (Rabs l * C); split; auto; split.
apply Rle_lt_trans with (Rabs l * 0); [ right; ring | idtac ].
apply Rmult_lt_compat_l; auto with real.
apply Rabs_pos_lt; auto.
intros X dX H5 H6;
  rewrite Rabs_mult;
  rewrite Rmult_assoc;
  apply Rmult_le_compat_l.
apply Rabs_pos.
apply H3; auto.
Qed.

(** Many other properties *)

Lemma OuP_multiplyAA :
  forall P (f:A-> R*R->R) g l,
  OuP P f g ->
  OuP P (scalar_multiplyAA_fun _ _ l f) g.
Proof.
unfold OuP, scalar_multiplyAA_fun; intros P f g l (alp, (C, (H1, (H2, H3)))).
case (Req_EM_T l 0); intros H4.
exists alp; exists C.
repeat split; auto; intros;
  rewrite H4;
  ring_simplify (0 * f X dX);
  rewrite Rabs_R0;
  apply Rle_trans with (Rabs (f X dX)).
apply Rabs_pos.
apply H3; auto.
exists alp; exists (Rabs l * C); split; auto; split.
apply Rle_lt_trans with (Rabs l * 0); [ right; ring | idtac ].
apply Rmult_lt_compat_l; auto with real.
apply Rabs_pos_lt; auto.
intros X dX H5 H6;
  rewrite Rabs_mult;
  rewrite Rmult_assoc;
  apply Rmult_le_compat_l.
apply Rabs_pos.
apply H3; auto.
Qed.

Lemma OuP_maj: 
  forall (P:R*R -> Prop) f1 f2 g, 
  (forall (X:A) (dX : R * R), P dX -> Rabs (f1 X dX) <= Rabs (f2 X dX)) ->
  OuP P f2 g ->
  OuP P f1 g.
intros.
elim H0; intros C (alp, (H1,(H2,H3))).
exists C; exists alp; split; trivial; split; trivial.
intros.
apply Rle_trans with (1:=H X dX H5).
now apply H3.
Qed.

Lemma OuP_inclusion :
  forall (P:R*R -> Prop) f g1 g2, 
  (forall X : R * R, norm_l2 X <= 1 -> P X -> Rabs (g1 X) <= Rabs (g2 X)) ->
  OuP P f g1 ->
  OuP P f g2.
unfold OuP; intros P f g1 g2 Hg [alp [C [H1 [H2 H3]]]].
exists (Rmin alp 1); exists C; repeat split; trivial.
apply Rmin_pos; auto with real.
intros a dX HdX1 Y.
apply Rle_trans with (C*Rabs (g1 dX)).
apply H3; auto.
apply Rlt_le_trans with (1:=HdX1); apply Rmin_l.
apply Rmult_le_compat_l; auto with real.
apply Hg; auto.
apply Rle_trans with (Rmin alp 1); auto with real; 
  apply Rmin_r.
Qed.

Lemma OuP_sum_ortho :
  forall P f1 f2, forall m k : nat,
  OuP P f1 (fun X => Rabs (fst X) ^ m) ->
  OuP P f2 (fun X => Rabs (snd X) ^ k) ->
  OuP P (fun a => add2_fun (f1 a) (f2 a)) 
            (fun X => Rabs (fst X) ^ m + Rabs (snd X) ^ k).
Proof.
intros P f1 f2 m k H1 H2.
apply OuP_is_EV_plus.
apply OuP_inclusion with (fun X : R * R => Rabs (fst X) ^ m); auto.
intros X H3 Y.
rewrite Rabs_right.
rewrite (Rabs_right (Rabs (fst X) ^ m + Rabs (snd X) ^ k)).
apply Rle_trans with (Rabs (fst X) ^ m + 0); [ right; ring | idtac ].
apply Rplus_le_compat_l.
apply pow_le; auto.
apply Rabs_pos.
apply Rle_ge.
apply Rle_trans with (0 + 0); auto with real.
apply Rplus_le_compat; apply pow_le; apply Rabs_pos.
apply Rle_ge; apply pow_le; apply Rabs_pos.
apply OuP_inclusion with (fun X : R * R => Rabs (snd X) ^ k); auto.            
intros X H3 Y.
rewrite Rabs_right.
rewrite (Rabs_right (Rabs (fst X) ^ m + Rabs (snd X) ^ k)).
apply Rle_trans with (0 + Rabs (snd X) ^ k); [ right; ring | idtac ].
apply Rplus_le_compat_r.
apply pow_le; auto.
apply Rabs_pos.
apply Rle_ge.
apply Rle_trans with (0 + 0); auto with real.
apply Rplus_le_compat; apply pow_le; apply Rabs_pos.
apply Rle_ge; apply pow_le; apply Rabs_pos.
Qed.

Lemma OuP_division :
  forall P f g1 g2,
  OuP P f g1 ->
  OuP P (fun a => divide2_fun (f a) g2) (divide2_fun g1 g2).
Proof.
unfold OuP; intros P f g1 g2 (alp, (C, (H2, (H3, H4)))).
exists alp; exists C; repeat split; auto.
intros; unfold divide2_fun, Rdiv; repeat rewrite Rabs_mult.
rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
apply Rabs_pos.
apply H4; auto.
Qed.

Lemma OuP_eq_trans_l :
  forall P:R*R-> Prop, forall f1 f2 g,
  (forall a dX, P dX -> f1 a dX = f2 a dX) ->
  OuP P f1 g-> OuP P f2 g.
intros P f1 f2 g H [alp [C [H1 [H2 H3]]]].
exists alp; exists C; intuition.
rewrite <- H; auto.
Qed.


Lemma OuP_eq_trans_r :
  forall (P:R*R-> Prop) f g1 g2,
  (forall dX : R * R, P dX -> g1 dX = g2 dX) ->
  OuP P f g1-> OuP P f g2.
intros P f g1 g2 H [alp [C [H1 [H2 H3]]]].
exists alp; exists C; intuition.
rewrite <- H; auto.
Qed.

Lemma OuP_implies_OuP: forall P1 P2: R*R -> Prop, forall f g,
  (forall dX, P2 dX -> P1 dX ) ->
  OuP P1 f g -> OuP P2 f g.
unfold OuP; intros P1 P2 f g H [alp [C [H1 [H2 H3]]]].
exists alp; exists C; intuition.
Qed.


Lemma OuP_implies_OuP1: forall P:R*R -> Prop, forall f g,
  (forall dX, P dX -> P (fst dX,0)) ->
  OuP P f g -> OuP P (fun X dX => f X (fst dX,0)) (fun dX => g (fst dX,0)).
unfold OuP; intros P f g H (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros X dX Y1 Y2.
apply H3; auto.
rewrite norm_l2_proj1.
now apply Rle_lt_trans with (1:=(Rabs_fst_le_norm_l2 dX)).
Qed.

Lemma OuP_implies_OuP2: forall P:R*R -> Prop, forall f g,
  (forall dX, P dX -> P (0, snd dX)) ->
  OuP P f g -> OuP P (fun X dX => f X (0, snd dX)) (fun dX => g (0,snd dX)).
unfold OuP; intros P f g H (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros X dX Y1 Y2.
apply H3; auto.
rewrite norm_l2_proj2.
now apply Rle_lt_trans with (1:=(Rabs_snd_le_norm_l2 dX)).
Qed.

Lemma OuP_implies_OuPMinus:
 forall P:R*R-> Prop, forall f g,
  (forall dX, P dX -> P (-fst dX, -snd dX)) ->
  OuP P f g -> OuP P (fun X dX => f X (-fst dX, -snd dX)) (fun dX => g (-fst dX, -snd dX)).
unfold OuP; intros P f g H (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros X dX Y1 Y2.
apply H3; auto.
apply Rle_lt_trans with (2:=Y1).
right; unfold norm_l2; simpl.
now repeat rewrite <- Rsqr_neg.
Qed.


Lemma OuP_multiply1: forall P:R*R -> Prop, forall f g,
  (forall dX, P dX -> fst dX <> 0) ->
  OuP P (fun X dX => (fst dX) * f X dX) 
      (fun dX => (fst dX) * g dX) 
    -> OuP P f g.
unfold OuP; intros P f g H (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros X dX Y1 Y2.
apply Rmult_le_reg_l with (Rabs (fst dX)).
apply Rabs_pos_lt.
now apply H.
rewrite <- Rabs_mult.
apply Rle_trans with (C * Rabs (fst dX * g dX)).
apply H3; auto.
rewrite Rabs_mult; right; ring.
Qed.

Lemma OuP_multiply2: forall P:R*R -> Prop, forall f g,
  (forall dX, P dX -> snd dX <> 0) ->
  OuP P (fun X dX => (snd dX) * f X dX) 
      (fun dX => (snd dX) * g dX) 
    -> OuP P f g.
unfold OuP; intros P f g H (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros X dX Y1 Y2.
apply Rmult_le_reg_l with (Rabs (snd dX)).
apply Rabs_pos_lt.
now apply H.
rewrite <- Rabs_mult.
apply Rle_trans with (C * Rabs (snd dX * g dX)).
apply H3; auto.
rewrite Rabs_mult; right; ring.
Qed.

Lemma OuP_divide1: forall P:R*R -> Prop, forall f g,
  (forall dX, P dX -> fst dX <> 0) ->
  OuP P (fun X dX => / (fst dX) * f X dX) 
      (fun dX => / (fst dX) * g dX) 
    -> OuP P f g.
unfold OuP; intros P f g H (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros X dX Y1 Y2.
apply Rmult_le_reg_l with (Rabs (/ (fst dX))).
apply Rabs_pos_lt.
auto with real.
rewrite <- Rabs_mult.
apply Rle_trans with (C * Rabs (/ fst dX * g dX)).
apply H3; auto.
rewrite Rabs_mult; right; ring.
Qed.

Lemma OuP_divide2: forall P:R*R -> Prop, forall f g,
  (forall dX, P dX -> snd dX <> 0) ->
  OuP P (fun X dX => / (snd dX) * f X dX) 
      (fun dX => / (snd dX) * g dX) 
    -> OuP P f g.
unfold OuP; intros P f g H (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros X dX Y1 Y2.
apply Rmult_le_reg_l with (Rabs (/ snd dX)).
apply Rabs_pos_lt.
auto with real.
rewrite <- Rabs_mult.
apply Rle_trans with (C * Rabs (/snd dX * g dX)).
apply H3; auto.
rewrite Rabs_mult; right; ring.
Qed.


Lemma OuP_dX_to_eps: forall P:R*R -> Prop, forall f g eps,
 { C:R | 0 < C /\ (forall X dX, norm_l2 (eps X dX) <= C * norm_l2 dX)} ->
 { C:R | 0 < C /\ (forall X dX, Rabs (g (eps X dX)) <= C * Rabs (g dX))} ->
 (forall X dX, P dX -> P (eps X dX)) ->
 OuP P f g -> OuP P (fun X dX => f X (eps X dX)) g.
unfold OuP; intros P f g eps (C1,(J1,J2)) (C2,(M1,M2)) K (alp, (C,(H1,(H2,H3)))).
exists (alp/C1); exists (C2*C).
split.
apply Rmult_lt_0_compat; auto with real.
split.
apply Rmult_lt_0_compat; auto.
intros.
apply Rle_trans with (C* (C2*Rabs (g dX)));[idtac|right; ring].
apply Rle_trans with (C* Rabs (g (eps X dX))).
apply H3.
apply Rle_lt_trans with (C1*norm_l2 dX).
apply J2.
apply Rmult_lt_reg_l with (/C1); auto with real.
apply Rle_lt_trans with (norm_l2 dX).
right; field; auto with real.
apply Rlt_le_trans with (1:= H); right; unfold Rdiv; ring.
now apply K.
apply Rmult_le_compat_l.
auto with real.
apply M2.
Qed.


Lemma OuP_X_to_appr: forall P:R*R -> Prop, forall f g appr,
 OuP P f g -> OuP P (fun X dX => f (appr X dX) dX) g.
unfold OuP; intros P f g appr (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
Qed.


Lemma OuP_snddX_to_dX: forall P:R*R -> Prop, forall h g,
    (forall dX, P dX -> P (0, snd dX))
   -> OuP P h g 
   -> OuP P (fun X dX => h X (0,snd dX)) (fun dX => g (0,snd dX)).
unfold OuP; intros P h g H (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros X dX H4 H5.
apply H3; auto.
rewrite norm_l2_proj2.
now apply Rle_lt_trans with (1:=Rabs_snd_le_norm_l2 dX).
Qed.

Lemma OuP_sqrt : forall P:R*R -> Prop, forall f g,
  (forall X dX, P dX -> 0 <= f X dX) ->
  OuP P f (fun X => Rsqr (g X)) -> OuP P (fun X dX => sqrt (f X dX)) g.
assert (forall r, Rsqr (Rabs r) = Rabs (Rsqr r)).
intros r; unfold Rsqr; now rewrite Rabs_mult.
unfold OuP; intros P f g K (alp,(C,(H1,(H2,H3)))).
exists alp; exists (sqrt C).
split; auto; split.
apply sqrt_lt_R0; auto.
intros.
apply Rsqr_incr_0_var.
rewrite Rsqr_mult.
rewrite Rsqr_sqrt; auto with real.
repeat rewrite H.
rewrite Rsqr_sqrt; auto.
apply Rmult_le_pos.
apply sqrt_pos.
apply Rabs_pos.
Qed.

Lemma OuP_Rsqr : forall P:R*R -> Prop, forall f g,
  (forall dX, P dX -> 0 <= g dX) ->
  OuP P f (fun dX => sqrt (g dX)) -> OuP P (fun X dX => Rsqr (f X dX)) g.
assert (forall r, Rsqr (Rabs r) = Rabs (Rsqr r)).
intros r; unfold Rsqr; now rewrite Rabs_mult.
unfold OuP; intros P f g K (alp,(C,(H1,(H2,H3)))).
exists alp; exists (Rsqr C).
split; auto; split.
apply Rlt_0_sqr; auto with real.
intros.
rewrite <- H.
rewrite <- (Rsqr_sqrt (Rabs (g dX))).
rewrite <- Rsqr_mult.
apply Rsqr_incr_1.
replace (sqrt (Rabs (g dX))) with (Rabs (sqrt (g dX))).
apply H3; auto.
repeat rewrite Rabs_right; try apply Rle_ge; auto with real.
apply sqrt_pos.
apply Rabs_pos.
apply Rmult_le_pos; auto with real.
apply sqrt_pos.
apply Rabs_pos.
Qed.


Lemma OuP_match : forall P i f0 f1 g,
  OuP P f0 g ->
  OuP P (fun X dX => f1 (S (pred (i X dX))) X dX) g ->
  OuP P (fun X dX => match (i X dX) with
     | 0%nat  => f0 X dX 
     | (S n) =>  f1 (S n) X dX
    end) 
  g.
unfold OuP; intros P i f0 f1 g (alp1,(C1,(A1,(A2,A3)))) (alp2,(C2,(B1,(B2,B3)))).
exists (Rmin alp1 alp2); exists (Rmax C1 C2).
split; try split.
now apply Rmin_pos.
apply Rlt_le_trans with (1:=A2); apply RmaxLess1.
intros X dX H1 H2.
case_eq (i X dX); intros.
apply Rle_trans with (C1* Rabs (g dX)).
apply A3; auto.
apply Rlt_le_trans with (1:=H1).
apply Rmin_l.
apply Rmult_le_compat_r.
apply Rabs_pos.
apply RmaxLess1.
rewrite <- H.
apply Rle_trans with (C2* Rabs (g dX)).
replace (i X dX) with (S (pred (i X dX))) by omega.
apply B3; auto.
apply Rlt_le_trans with (1:=H1).
apply Rmin_r.
apply Rmult_le_compat_r.
apply Rabs_pos.
apply RmaxLess2.
Qed.


Lemma OuP_Rabs: forall P f g,
  OuP P f g -> OuP P (fun x dx => Rabs (f x dx)) g.
intros  P f g (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; try split; trivial.
intros.
rewrite Rabs_Rabsolu.
now apply H3.
Qed.


End BigOuP.

Section BigOuP2.

Lemma OuP_X_to_appr_fst: forall P:R*R -> Prop, forall h g (appr: R*R -> R*R-> R*R),
  (forall x dX, snd (appr (x,0) dX) = 0)  
   -> OuP P (fun x dX => h (x,0) dX)  g 
   -> OuP P (fun (X:R*R) dX => h (appr (fst X, 0) dX) dX) g.
unfold OuP; intros P h g appr H (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros.
generalize (H (fst X) dX).
destruct (appr (fst X, 0) dX).
simpl; intros H'; rewrite H'.
apply H3; auto.
Qed.

Lemma OuP_X_to_x: forall P:R*R -> Prop, forall h g,
      OuP P (fun (X:R*R) dX => h (fst X) dX)  g 
   -> OuP P (fun x dX => h x dX) g.
unfold OuP; intros P h g (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; auto; split; auto.
intros x dX H4 H5.
replace x with (fst (x,0)) by now simpl.
apply H3; auto.
Qed.



End BigOuP2.

(** Now, we consider for A the interval between 0 and T *)

Definition Under (T:R) := fun (r:R) => (0 <= r <= T)%R.

Lemma OuP_sum_f: forall (P:R*R-> Prop) (T:R) f g,
   0 < T
 ->  (forall dX, P dX -> (0 < snd dX)%R)
 ->  OuP P (fun (t: {t| Under T t}) dX =>  f dX (kt (projT1 t) (snd dX)))  g
 ->  OuP P (fun (t: {t | Under T t}) dX =>  snd dX * 
            sum_f 1 (kt (projT1 t) (snd dX)) (fun n : nat => f dX n)) g.
intros P T f g M1 M2.
intros (alp, (C,(H1,(H2,H3)))).
exists (Rmin T alp); exists (C*2*T).
split.
now apply Rmin_pos.
split.
repeat apply Rmult_lt_0_compat; auto with real.
intros (t,Ht) dX H4 H5; simpl.
rewrite Rabs_mult.
apply Rle_trans with 
 (Rabs (snd dX) * ((INR (kt t (snd dX)-1)+1) * (C * Rabs (g dX)))).
apply Rmult_le_compat_l.
apply Rabs_pos.
apply Rle_trans with (sum_f 1 (kt t (snd dX)) (fun n : nat => Rabs (f dX n))).
apply sum_f_triang.
apply sum_f_le.
intros.
assert (K: Under T (INR i*snd dX)).
split.
apply Rmult_le_pos.
auto with real.
left; now apply M2.
apply Rle_trans with (INR (kt t (snd dX)) * snd dX).
apply Rmult_le_compat_r.
left; now apply M2.
destruct H; auto with real.
apply Rle_trans with ((t/snd dX)*snd dX).
apply Rmult_le_compat_r.
left; now apply M2.
apply kt_le.
eapply Ht.
now apply M2.
apply Rle_trans with t.
right; field.
apply sym_not_eq; apply Rlt_not_eq.
now apply M2.
eapply Ht.
replace i with (kt (projT1 (existT (Under T) (INR i * snd dX)  K)) (snd dX)).
apply H3 with (dX:=dX) (X:=(existT (Under T) (INR i * snd dX) K)); trivial.
apply Rlt_le_trans with (1:=H4).
apply Rmin_r.
simpl.
apply kt_eq.
apply sym_not_eq; apply Rlt_not_eq.
now apply M2.
apply Rmult_le_pos. 
now left.
apply Rabs_pos.
replace (C * 2 * T * Rabs (g dX)) with ((2*T)*(C*Rabs (g dX))) by ring.
rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
apply Rmult_le_pos. 
now left.
apply Rabs_pos.
rewrite Rabs_right.
apply Rle_trans with (snd dX * (t/snd dX + 1)).
apply Rmult_le_compat_l.
left; now apply M2.
apply Rplus_le_compat_r.
apply Rle_trans with (INR (kt t (snd dX))).
apply le_INR.
auto with zarith.
apply kt_le.
eapply Ht.
now apply M2.
rewrite Rmult_plus_distr_l.
replace (2*T) with (T+T) by ring.
apply Rplus_le_compat.
apply Rle_trans with t.
right; field.
apply sym_not_eq; apply Rlt_not_eq.
now apply M2.
eapply Ht.
rewrite Rmult_1_r.
rewrite <- (Rabs_right (snd dX)).
apply Rle_trans with (norm_l2 dX).
apply Rabs_snd_le_norm_l2.
left; apply Rlt_le_trans with (1:=H4).
apply Rmin_l.
apply Rle_ge; left; now apply M2.
apply Rle_ge; left; now apply M2.
Qed.

Lemma OuP_sum_f_minus1_bis: forall (P:R*R-> Prop) (T:R) f g,
   0 < T
 ->  (forall dX, P dX -> (0 < snd dX)%R)
 ->  OuP P (fun (t: {t| Under T t}) dX =>  f dX (kt (projT1 t) (snd dX)))  g
 ->  OuP P (fun (t: {t | Under T t}) dX =>  snd dX * 
            sum_f 1 (kt (projT1 t) (snd dX)-1) (fun n : nat => f dX (S n))) g.
intros.
apply OuP_maj with (fun (t : {t : R | Under T t}) =>
  add2_fun 
    (fun dX => Rabs (snd dX * sum_f 1 (kt (projT1 t) (snd dX)) (fun n : nat => f dX n)))
    (scalar_multiplyA_fun _ 2 (fun dX => Rabs (snd dX *f dX 1%nat)))).
intros; unfold add2_fun, scalar_multiplyA_fun.
rewrite (Rabs_right ((Rabs (snd dX * sum_f 1 (kt (projT1 X) (snd dX)) (fun n : nat => f dX n)) +
   (2*Rabs (snd dX * f dX 1%nat))))).
repeat rewrite Rabs_mult.
replace (2 * (Rabs (snd dX) * Rabs (f dX 1%nat))) with
    (Rabs (snd dX) * Rabs (f dX 1%nat) + Rabs (snd dX) * Rabs (f dX 1%nat)) by ring.
repeat rewrite <- Rmult_plus_distr_l.
apply Rmult_le_compat_l.
apply Rabs_pos.
rewrite <- Rplus_assoc.
apply sum_f_S_Rabs_minus1.
apply Rle_ge; replace 0 with (0+(2*0)) by ring.
apply Rplus_le_compat.
apply Rabs_pos.
apply Rmult_le_compat_l; try apply Rabs_pos; auto with real.
apply  OuP_is_EV_plus.
apply OuP_Rabs.
apply OuP_sum_f; trivial.
apply OuP_is_EV_mult.
apply OuP_Rabs.
destruct H1 as (alp,(C,(H1,(H2,H3)))).
exists (Rmin (Rmin 1 T) alp); exists C; split.
repeat apply Rmin_pos; auto with real.
split; trivial.
intros.
rewrite Rabs_mult.
replace (C*Rabs (g dX)) with (1*(C*Rabs (g dX))) by ring.
apply Rmult_le_compat.
apply Rabs_pos.
apply Rabs_pos.
apply Rle_trans with (Rmin (Rmin 1 T) alp).
left; apply Rle_lt_trans with (2:=H4).
apply Rabs_snd_le_norm_l2.
apply Rle_trans with (2:=Rmin_l 1 T); apply Rmin_l.
assert (L: 0 <= snd dX <= T).
split.
left; now apply H0.
rewrite <- (Rabs_right (snd dX)).
apply Rle_trans with (1:=Rabs_snd_le_norm_l2 dX).
left; apply Rlt_le_trans with (1:=H4).
apply Rle_trans with (2:=Rmin_r 1 T); apply Rmin_l.
apply Rle_ge; left; now apply H0.
replace (1%nat) with (kt (projT1 (existT (Under T) (snd dX) L)) (snd dX)).
apply H3 with (X:=(existT (Under T) (snd dX) L)); trivial.
apply Rlt_le_trans with (1:=H4).
apply Rmin_r.
unfold kt; simpl.
replace (snd dX/snd dX) with (IZR 1).
rewrite Int_part_IZR; reflexivity.
unfold IZR; simpl; field.
apply Rgt_not_eq; now apply H0.
Qed.


Lemma OuP_every_implies_OuP_here:
     forall P Q f g, 
      OuP P (fun (t:R) (dX:R*R) => f t dX) g
  -> OuP P (fun (t:{t | Q t}) (dX:R*R) => f (projT1 t) dX) g.
intros P Q f g (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; trivial; split; trivial.
intros.
apply H3; auto.
Qed.

Lemma OuP_unit: forall P T f g, 
     OuP P (fun (_:unit) dX => f dX) g
  -> OuP P (fun (_:T)    dX => f dX) g.
intros  P T f g (alp,(C,(H1,(H2,H3)))).
exists alp; exists C; split; trivial; split; trivial.
intros.
apply H3; intuition.
Qed.

(** A few notations *)

Notation Ou   := (OuP (fun _ => True)).
Notation Oup  := (OuP (fun dX => 0 <= fst dX /\ 0 <= snd dX)).
Notation Oups := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX)).


