(** Differentiability, Taylor expansions *)

Require Export BigO.
Open Local Scope nat_scope.
Open Local Scope R_scope.

Definition Xjk (dX : R * R) (j: Z) (k:nat) : R * R :=
  ((IZR j)*(fst dX), (INR k)*(snd dX)).

Definition Xxt (X dX:R*R) : R*R :=
  Xjk dX (jx (fst X) (fst dX)) (kt (snd X) (snd dX)).


(** Differentiability *)

Parameter partial_derive_firstvar: (R * R -> R) -> (R * R -> R).
Parameter partial_derive_secondvar: (R * R -> R) -> (R * R -> R).

Fixpoint partial_derive_firstvar_n (m : nat) (f : R * R -> R) : R * R -> R :=
  match m with
  | 0 => f
  | S m' => partial_derive_firstvar (partial_derive_firstvar_n m' f)
  end.

Fixpoint partial_derive_secondvar_n (k : nat) (f : R * R -> R) : R * R -> R :=
  match k with
  | 0 => f
  | S k' => partial_derive_secondvar (partial_derive_secondvar_n k' f)
  end.

Definition partial_derive (m k : nat) (f : R * R -> R) : R * R -> R :=
  partial_derive_firstvar_n m (partial_derive_secondvar_n k f).

(** Taylor expansions *)

Definition differential (p : nat) (f : R * R -> R) (X dX : R * R) : R :=
  sum_f_R0 
    (fun m =>
      C p m *
      partial_derive m (p - m)%nat f X *
      fst dX ^ m * snd dX ^ (p - m)%nat)
    p.

Definition DL_pol (n : nat) (f : R * R -> R) (X dX : R * R) : R :=
  sum_f_R0
    (fun p =>
      differential p f X dX / INR (fact p))
    n.

(** Definition of being sufficiently regular by the Taylor expansions *)

Definition is_sufficiently_regular_n n f :=
  forall m, (m <= n)%nat ->
  Ou (fun X => 
     subtract2_fun
      (fun dX => f (add2 X dX))
      (fun dX => DL_pol (m - 1)%nat f X dX))
    (fun dX => norm_l2 dX ^ m).

(** A few lemmas needed by the consistancy (many computations inside) *)

Lemma DL_discretization_eps_aux:   forall f : R * R -> R, forall eps: R * R -> R * R -> R * R,
  (is_sufficiently_regular_n 4 f) ->
  { C : R |
     0 < C /\ (forall X dX : R * R, norm_l2 (eps X dX) <= C * norm_l2 dX)} ->
  Ou ((fun X dX => f (add2 X (eps X dX)) - 2*f X + f (subtract2 X (eps X dX)) 
           - (fst (eps X dX))^2 * (partial_derive_firstvar_n 2 f X)
           - (snd (eps X dX))^2 * (partial_derive_secondvar_n 2 f X)
           - 2*(fst (eps X dX))*(snd (eps X dX)) * partial_derive_firstvar (partial_derive_secondvar f) X
            ))
     (fun dX => (norm_l2 dX)^4).
intros f eps H (C', (J1,J2)).
(* c'est exactement la somme de 2 DL *)
apply OuP_eq_trans_l with (fun X => add2_fun
    (fun dX => f (add2 X (eps X dX))  - DL_pol (4 - 1)%nat f X (eps X dX))
    (fun dX => f (add2 X (scalar_multiply2 (-1) (eps X dX))) 
           - DL_pol (4 - 1)%nat f X (scalar_multiply2 (-1) (eps X dX)))).
intros; unfold add2_fun; unfold DL_pol,scalar_multiply2, subtract2, add2 ; simpl.
unfold differential; simpl.
unfold partial_derive; simpl.
unfold C; simpl.
replace (snd a + -1 * snd (eps a dX)) with (snd a - snd (eps a dX)) by ring.
replace (fst a + -1 * fst (eps a dX)) with (fst a - fst (eps a dX)) by ring.
field.
apply OuP_is_EV_plus.
(* DL normal *)
apply OuP_dX_to_eps with (eps:=eps) (f:= fun X dX => f (X+dX)%type - DL_pol (4-1) f X dX).
exists C'; split; auto with real.
exists (C'*C'*C'*C'); split.
now repeat apply Rmult_lt_0_compat.
intros; repeat rewrite Rabs_right.
apply Rle_trans with ((C' * norm_l2 dX)^4).
apply pow_incr.
split; auto.
apply norm_l2_pos.
right; unfold pow; ring.
apply Rle_ge; apply pow_le; apply norm_l2_pos.
apply Rle_ge; apply pow_le; apply norm_l2_pos.
intros; auto.
now apply OuP_eq_trans_l with (2 := H _ (le_refl 4)).
(* DL avec -eps X dX *)
apply OuP_dX_to_eps with (eps:=fun X dX => scalar_multiply2 (-1) (eps X dX)) 
     (f:= fun X dX => f (X+dX)%type - DL_pol (4-1) f X dX).
3: easy.
exists C'; split; auto with real.
intros; rewrite norm_l2_scalar_mult.
rewrite Rabs_Ropp; rewrite Rabs_R1; rewrite Rmult_1_l.
apply J2.
exists (C'*C'*C'*C'); split.
now repeat apply Rmult_lt_0_compat.
intros; repeat rewrite Rabs_right.
rewrite norm_l2_scalar_mult.
rewrite Rabs_Ropp; rewrite Rabs_R1; rewrite Rmult_1_l.
apply Rle_trans with ((C' * norm_l2 dX)^4).
apply pow_incr.
split; auto.
apply norm_l2_pos.
right; unfold pow; ring.
apply Rle_ge; apply pow_le; apply norm_l2_pos.
apply Rle_ge; apply pow_le; apply norm_l2_pos.
now apply OuP_eq_trans_l with (2 := H _ (le_refl 4)).
Qed.



Lemma DL_discretization_eps_dx:   forall f : R * R -> R, 
  (is_sufficiently_regular_n 4 f) ->
  Ou ((fun X dX => f (add2 X (fst dX, 0)) - 2*f X + f (subtract2 X (fst dX,0)) 
           - (fst dX)^2 * (partial_derive_firstvar_n 2 f X)))
     (fun dX => (fst dX)^4).
intros.
apply OuP_eq_trans_r with (fun dX => norm_l2 (fst dX,0) ^4).
intros; rewrite norm_l2_proj1.
rewrite RPow_abs.
apply Rabs_right.
apply Rle_ge; replace ((fst dX) ^ 4) with (Rsqr (Rsqr (fst dX))); auto with real.
unfold Rsqr, pow; ring.
apply OuP_eq_trans_l with (fun X dX => (f (X + (fst dX, 0))%type - 2 * f X + f (X - (fst dX, 0))%type 
   - (fst (fst dX,0)) ^ 2 * partial_derive_firstvar_n 2 f X
   - (snd (fst dX,0))^2 * (partial_derive_secondvar_n 2 f X)
   - 2*(fst (fst dX,0))*(snd (fst dX,0)) * partial_derive_firstvar (partial_derive_secondvar f) X)).
intros; simpl; ring.
apply OuP_implies_OuP1 with 
  (f:=(fun X dX => f (add2 X dX) - 2*f X + f (subtract2 X dX) 
           - (fst dX)^2 * (partial_derive_firstvar_n 2 f X)
           - (snd dX)^2 * (partial_derive_secondvar_n 2 f X)
           - 2*(fst dX)*(snd dX) * partial_derive_firstvar (partial_derive_secondvar f) X)) 
  (g:= fun dX => (norm_l2 dX)^4).
intros; auto.
apply DL_discretization_eps_aux; auto.
exists 1; split; auto with real.
Qed.


Lemma DL_discretization_eps_dt:   forall f : R * R -> R, 
  (is_sufficiently_regular_n 4 f) ->
  Ou ((fun X dX => f (add2 X (0, snd dX)) - 2*f X + f (subtract2 X (0, snd dX)) 
           - (snd dX)^2 * (partial_derive_secondvar_n 2 f X)))
     (fun dX => (snd dX)^4).
intros.
apply OuP_eq_trans_r with (fun dX => norm_l2 (0, snd dX) ^4).
intros; rewrite norm_l2_proj2.
rewrite RPow_abs.
apply Rabs_right.
apply Rle_ge; replace ((snd dX) ^ 4) with (Rsqr (Rsqr (snd dX))); auto with real.
unfold Rsqr, pow; ring.
apply OuP_eq_trans_l with (fun X dX => (f (X + (0, snd dX))%type - 2 * f X + f (X - (0,snd dX))%type 
   - (fst (0,snd dX)) ^ 2 * partial_derive_firstvar_n 2 f X
   - (snd (0,snd dX))^2 * (partial_derive_secondvar_n 2 f X)
   - 2*(fst (0, snd dX))*(snd (0,snd dX)) * partial_derive_firstvar (partial_derive_secondvar f) X)).
intros; simpl; ring.
apply OuP_implies_OuP2 with 
  (f:=(fun X dX => f (add2 X dX) - 2*f X + f (subtract2 X dX) 
           - (fst dX)^2 * (partial_derive_firstvar_n 2 f X)
           - (snd dX)^2 * (partial_derive_secondvar_n 2 f X)
           - 2*(fst dX)*(snd dX) * partial_derive_firstvar (partial_derive_secondvar f) X)) 
  (g:= fun dX => (norm_l2 dX)^4).
intros; auto.
apply DL_discretization_eps_aux; auto.
exists 1; split; auto with real.
Qed.



