(** R^n *)

Require Export Differential.
Require Export Fourier.
Require Export Reals_compl.
Require ZArith.
Open Local Scope R_scope.


(** R^n is a vector space.
  Vectors in R^n are represented by nat -> R functions *)

Definition add_n (f g : Z -> R) : Z -> R :=
  fun i => f i + g i.

Definition subtract_n (f g : Z -> R) : Z -> R :=
  fun i => f i - g i.

Definition scalar_multiply_n (l : R) (f : Z -> R) : Z -> R :=
  fun i => l * f i.

Notation "f + g" := (add_n f g) : type_scope.
Notation "f - g" := (subtract_n f g) : type_scope.

(** Definition of the dot_product on finite sets *)

Fixpoint dot_product_n (n:nat) (i:Z) (f g : Z -> R) {struct n} : R :=
  match n with
  | 0 => f i * g i
  | S n' => (f i) * (g i) + dot_product_n n' (i+1) f g
  end.

Definition dot_product_z (n1 n2:Z) (f g : Z -> R) : R :=
       dot_product_n (Zabs_nat (n2-n1)) n1 f g.

Definition squared_norm_z (n1 n2 : Z) (f : Z -> R) := dot_product_z n1 n2 f f.

(** Many lemmas *)

Lemma dot_product_n_zero :
  forall u v n a,
    (forall i : Z, 
        (a <= i <= a + Z_of_nat n)%Z -> u i = 0 \/ v i = 0) ->
    dot_product_n n a u v = 0.
Proof.
  induction n; intuition.
  simpl.
  destruct (H a); intuition.
  simpl.
  destruct (H a); intuition.
  rewrite H0; rewrite IHn; intuition. ring.
  apply H; intuition. rewrite inj_S; omega.
  rewrite H0; rewrite IHn; intuition. ring.
  apply H; intuition. rewrite inj_S; omega.
Qed.

Lemma dot_product_n_split :
  forall u v n1 n2 a,
  dot_product_n (n1 + S n2) a u v = 
  dot_product_n n1 a u v + dot_product_n n2 (a+Z_of_nat n1+1)%Z u v.
Proof.
  induction n1.
  simpl; intros. 
  replace (a+0+1)%Z with (a+1)%Z by omega.
  reflexivity.
  intros.
  replace (a + (Z_of_nat (S n1)) + 1)%Z with ((a+1) + Z_of_nat n1 + 1)%Z.
  simpl.
  rewrite IHn1.
  ring.
  rewrite inj_S.
  omega.
Qed.

Lemma dot_product_z_zero_left:
  forall u v a a' b, (a' <= a <= b)%Z ->
  (forall i, (a' <= i < a)%Z -> u i = 0 \/ v i = 0) -> 
  dot_product_z a b u v =  dot_product_z a' b u v.
Proof.
  unfold dot_product_z; intros.
  assert (h: (a' = a \/ a' < a)%Z) by omega. destruct h.
  rewrite H1; reflexivity.
  pose (n1 := Zabs_nat (a - a' - 1)).
  pose (n2 := Zabs_nat (b - a)).
  replace (Zabs_nat (b - a')) with (n1 + S n2)%nat.
  rewrite dot_product_n_split.
  rewrite dot_product_n_zero with (a := a').
  replace (a' + Z_of_nat n1 + 1)%Z with a.
  unfold n2; ring.
  unfold n1.
  rewrite inj_Zabs_nat.
  rewrite Zabs_eq; omega.
  unfold n1. rewrite inj_Zabs_nat. rewrite Zabs_eq.
  intros; apply H0; try omega.
  omega.
  unfold n1; unfold n2.
  rewrite <- Zabs_nat_Zsucc; try omega.
  rewrite <- Zabs_nat_Zplus; try omega.
  replace (a - a' - 1 + Zsucc (b - a))%Z with (b-a')%Z by omega.
  reflexivity.
Qed.

Lemma dot_product_n_zero_right:
  forall u v n1 n2 a, 
  (forall i, (a + Z_of_nat n1 < i <= a + Z_of_nat (n1+n2))%Z -> u i = 0 \/ v i = 0) -> 
  dot_product_n (n1+n2)%nat a u v =  dot_product_n n1 a u v.
Proof.
  destruct n2.
  intros; replace (n1+0)%nat with n1 by omega.
  reflexivity.
  intros.
  rewrite dot_product_n_split.
  rewrite dot_product_n_zero with (n := n2).
  ring.
  intros; apply H.
  destruct H0.
  split; try omega. 
  rewrite inj_plus; rewrite inj_S; omega.
Qed.

Lemma dot_product_z_zero_right:
  forall u v a b b', (a <= b <= b')%Z ->
  (forall i, (b < i <= b')%Z -> u i = 0 \/ v i = 0) -> 
  dot_product_z a b u v =  dot_product_z a b' u v.
Proof.
  unfold dot_product_z; intros.
  replace  (Zabs_nat (b' - a)) with (Zabs_nat (b - a) + Zabs_nat (b' - b))%nat.
  rewrite dot_product_n_zero_right.
  reflexivity.
  intro i. rewrite inj_plus.
  repeat rewrite inj_Zabs_nat.
  repeat rewrite Zabs_eq; try omega.
  intro; apply H0; omega.
  rewrite <- Zabs_nat_Zplus; try omega.
  replace (b-a+(b'-b))%Z with (b'-a)%Z by omega.
  reflexivity.
Qed.

Lemma dot_product_z_zero :
  forall u v a b, (forall i, u i = 0 \/ v i = 0) -> dot_product_z a b u v = 0.
Proof.
  intros; unfold dot_product_z.
  apply dot_product_n_zero.
  auto.
Qed.

Lemma dot_product_n_distr_l :
  forall (n : nat) (i:Z)  (f g h : Z -> R), 
  dot_product_n n i f g + dot_product_n n i h g =
  dot_product_n n i (add_n f h) g.
intros n; induction n.
intros;unfold add_n;  simpl; ring.
intros; simpl; rewrite <- IHn; unfold add_n; simpl; ring.
Qed.

Lemma dot_product_z_distr_l :
  forall (n1 n2 : Z) (f g h : Z -> R), 
  dot_product_z n1 n2 f g + dot_product_z n1 n2 h g =
  dot_product_z n1 n2 (add_n f h) g.
intros; unfold dot_product_z.
apply dot_product_n_distr_l.
Qed.


Lemma dot_product_n_distr_r :
  forall (n : nat) (i:Z)  (f g h : Z -> R), 
  dot_product_n n i f g + dot_product_n n i f h =
  dot_product_n n i f (add_n g h).
intros n; induction n.
intros; unfold add_n; simpl; auto with real.
intros; simpl; rewrite <- IHn; unfold add_n; simpl; ring.
Qed.

Lemma dot_product_z_distr_r :
  forall (n1 n2 : Z) (f g h : Z -> R), 
  dot_product_z n1 n2 f g + dot_product_z n1 n2 f h =
  dot_product_z n1 n2 f (add_n g h).
intros; unfold dot_product_z.
apply dot_product_n_distr_r.
Qed.

Lemma dot_product_n_sub_distr_l :
  forall (n : nat) (i:Z)  (f g h : Z -> R), 
  dot_product_n n i f g - dot_product_n n i h g =
  dot_product_n n i (subtract_n f h) g.
intros n; induction n.
intros;unfold subtract_n;  simpl; ring.
intros; simpl; rewrite <- IHn; unfold subtract_n; simpl; ring.
Qed.

Lemma dot_product_z_sub_distr_l :
  forall (n1 n2 : Z) (f g h : Z -> R), 
  dot_product_z n1 n2 f g - dot_product_z n1 n2 h g =
  dot_product_z n1 n2 (subtract_n f h) g.
intros; unfold dot_product_z.
apply dot_product_n_sub_distr_l.
Qed.


Lemma dot_product_n_sub_distr_r :
  forall (n : nat) (i:Z)  (f g h : Z -> R), 
  dot_product_n n i f g - dot_product_n n i f h =
  dot_product_n n i f (subtract_n g h).
intros n; induction n.
intros; unfold subtract_n; simpl; ring.
intros; simpl; rewrite <- IHn; unfold subtract_n; simpl; ring.
Qed.

Lemma dot_product_z_sub_distr_r :
  forall (n1 n2 : Z) (f g h : Z -> R), 
  dot_product_z n1 n2 f g - dot_product_z n1 n2 f h =
  dot_product_z n1 n2 f (subtract_n g h).
intros; unfold dot_product_z.
apply dot_product_n_sub_distr_r.
Qed.

Lemma dot_product_n_scal_mult_l :
  forall (n : nat) (i:Z) (l : R) (f g : Z -> R),
  dot_product_n n i (scalar_multiply_n l f) g =
  l * dot_product_n n i f g.
intros n; induction n.
intros; unfold scalar_multiply_n; simpl; auto with real.
intros; simpl; rewrite Rmult_plus_distr_l.
rewrite <- IHn; unfold scalar_multiply_n; simpl; ring.
Qed.


Lemma dot_product_z_scal_mult_l :
  forall (n1 n2:Z) (l : R) (f g : Z -> R),
  dot_product_z n1 n2 (scalar_multiply_n l f) g =
  l * dot_product_z n1 n2 f g.
intros; unfold dot_product_z.
apply dot_product_n_scal_mult_l.
Qed.

Lemma dot_product_n_scal_mult_r :
  forall (n : nat) (i:Z) (l : R) (f g : Z -> R),
  dot_product_n n i f (scalar_multiply_n l g) =
  l * dot_product_n n i f g.
intros n; induction n.
intros; unfold scalar_multiply_n; simpl; ring.
intros; simpl; rewrite Rmult_plus_distr_l.
rewrite <- IHn; unfold scalar_multiply_n; simpl; ring.
Qed.


Lemma dot_product_z_scal_mult_r :
  forall (n1 n2:Z) (l : R) (f g : Z -> R),
  dot_product_z n1 n2 f (scalar_multiply_n l g) =
  l * dot_product_z n1 n2 f g.
intros; unfold dot_product_z.
apply dot_product_n_scal_mult_r.
Qed.


Lemma dot_product_n_sym :
  forall (n : nat) (i:Z) (f g : Z -> R),
  dot_product_n n i f g = dot_product_n n i g f.
intros n; induction n.
simpl; auto with real.
intros; simpl; rewrite <- IHn; simpl; ring.
Qed.

Lemma dot_product_z_sym :
  forall (n1 n2:Z) (f g : Z -> R),
  dot_product_z n1 n2 f g = dot_product_z n1 n2 g f.
intros; unfold dot_product_z.
apply dot_product_n_sym.
Qed.


Lemma dot_product_n_eq_suf_l :
  forall (n : nat) (i:Z) (f g h: Z -> R),
  (forall j : Z, (i <= j <= i+Z_of_nat n)%Z -> f j = g j) ->
   dot_product_n n i f h = dot_product_n n i g h.
intros n; induction n.
intros; simpl; auto with real.
rewrite H; auto with zarith.
intros; simpl; rewrite IHn with (g:=g); auto with real zarith.
rewrite H; auto with real zarith.
intros; apply H; auto with zarith.
rewrite inj_S; auto with zarith.
Qed.

Lemma dot_product_z_eq_suf_l :
  forall (n1 n2:Z) (f g h: Z -> R),
   (n1 <= n2)%Z ->
  (forall j : Z, (n1 <= j <= n2)%Z -> f j = g j) ->
   dot_product_z n1 n2 f h = dot_product_z n1 n2 g h.
intros; apply dot_product_n_eq_suf_l.
intros; apply H0.
destruct H1 as (T1,T2).
split; auto; apply Zle_trans with (1:=T2).
rewrite inj_Zabs_nat.
rewrite Zabs_eq; auto with zarith.
Qed.

Lemma dot_product_n_eq_suf_r :
  forall (n : nat) (i:Z) (f g h: Z -> R),
  (forall j : Z, (i <= j <= i+Z_of_nat n)%Z -> f j = g j) ->
   dot_product_n n i h f = dot_product_n n i h g.
intros n; induction n.
intros; simpl; auto with real.
rewrite H; auto with zarith.
intros; simpl; rewrite IHn with (g:=g); auto with real zarith.
rewrite H; auto with real zarith.
intros; apply H; auto with zarith.
rewrite inj_S; auto with zarith.
Qed.

Lemma dot_product_z_eq_suf_r :
  forall (n1 n2:Z) (f g h: Z -> R),
   (n1 <= n2)%Z ->
  (forall j : Z, (n1 <= j <= n2)%Z -> f j = g j) ->
   dot_product_z n1 n2 h f = dot_product_z n1 n2 h g.
intros; apply dot_product_n_eq_suf_r.
intros; apply H0.
destruct H1 as (T1,T2).
split; auto; apply Zle_trans with (1:=T2).
rewrite inj_Zabs_nat.
rewrite Zabs_eq; auto with zarith.
Qed.

Lemma dot_product_n_shift_indices: forall (n:nat) (i:Z) (f1 f2 : Z -> R),
  dot_product_n n i f1 f2 = 
        dot_product_n n i (fun j => f1 (j+1)%Z) (fun j => f2 (j+1)%Z) 
         + f1 i*f2 i - f1 (i+Z_of_nat n+1)%Z * f2 (i+Z_of_nat n+1)%Z.
intros n; induction n.
intros; simpl.
replace (i + 0+1)%Z with (i+1)%Z by omega.
ring.
intros; rewrite inj_S.
simpl; rewrite IHn.
replace (i + Zsucc (Z_of_nat n) + 1)%Z with (i + 1 + Z_of_nat n + 1)%Z by omega.
ring.
Qed.

Lemma dot_product_z_shift_indices: forall (n1 n2:Z) (f1 f2 : Z -> R),
  (n1 <= n2)%Z ->
  dot_product_z n1 n2 f1 f2 = 
        dot_product_z n1 n2 (fun j => f1 (j+1)%Z) (fun j => f2 (j+1)%Z) 
         + f1 n1*f2 n1 - f1 (n2+1)%Z * f2 (n2+1)%Z.
intros; unfold dot_product_z.
rewrite dot_product_n_shift_indices.
replace (n1 + Z_of_nat (Zabs_nat (n2 - n1)) + 1)%Z with (n2 + 1)%Z.
ring.
rewrite inj_Zabs_nat.
rewrite Zabs_eq; auto with zarith.
Qed.

Lemma squared_norm_z_eq: forall (n1 n2 : Z) (f g: Z -> R),
   (n1 <= n2)%Z ->
  (forall i : Z, (n1 <= i <= n2)%Z -> f i = g i) ->
  squared_norm_z n1 n2 f = squared_norm_z n1 n2 g.
intros; unfold squared_norm_z.
rewrite dot_product_z_eq_suf_l with n1 n2 f g f; auto.
rewrite dot_product_z_eq_suf_r with n1 n2 f g g; auto.
Qed.

Lemma squared_norm_z_pos_aux:
  forall (n:nat) (i: Z) (f: Z -> R),
    0 <= (dot_product_n n i f f).
intros n; induction n.
intros; simpl; fold (Rsqr (f i)); auto with real.
intros; simpl.
apply Rle_trans with (0+0); try apply Rplus_le_compat; auto with real.
fold (Rsqr (f i)); auto with real.
Qed.

Lemma squared_norm_z_pos:
  forall (n1 n2: Z) (f: Z -> R),
     (0 <= squared_norm_z n1 n2 f).
intros; unfold squared_norm_z; unfold dot_product_z.
apply squared_norm_z_pos_aux.
Qed.


Lemma squared_norm_n_scal_mult_n:  forall 
  (n1 n2:Z) (r:R) (f:Z -> R),
   (squared_norm_z n1 n2 (scalar_multiply_n r f)) = r*r*(squared_norm_z n1 n2 f).
intros; unfold squared_norm_z.
rewrite dot_product_z_scal_mult_l; rewrite dot_product_z_scal_mult_r; auto with real.
Qed.

Lemma squared_norm_n_plus:  forall 
  (n1 n2:Z) (f g:Z->R),
  (squared_norm_z n1 n2 (f+g)%type = squared_norm_z n1 n2 f + squared_norm_z n1 n2 g 
  + 2 * dot_product_z n1 n2 f g)%R.
intros n1 n2 f g; unfold squared_norm_z.
rewrite <- dot_product_z_distr_l.
repeat rewrite <- dot_product_z_distr_r.
rewrite <- dot_product_z_sym with (f:=f) (g:=g); ring.
Qed.


Lemma squared_norm_z_majoration_aux: forall n i f M,
  (forall j,  (i <= j <= i+Z_of_nat n)%Z -> Rsqr (f j) <= M) ->
  dot_product_n n i f f <= INR (S n)  * M.
intros n; induction n.
unfold Rsqr; simpl; intros.
rewrite Rmult_1_l; apply H; auto with zarith.
intros; simpl (dot_product_n (S n) i f f).
apply Rle_trans with (M + (INR (S n) *M)).
apply Rplus_le_compat.
unfold Rsqr in H; apply H; auto with zarith.
apply IHn.
intros; apply H; split; auto with zarith.
rewrite inj_S; auto with zarith.
repeat rewrite S_INR; simpl; right; ring. 
Qed.

Lemma squared_norm_z_majoration: forall n1 n2 f M,
  (n1 <= n2)%Z ->
  (forall j,  (n1 <= j <= n2)%Z -> Rsqr (f j) <= M) ->
  squared_norm_z n1 n2 f <= IZR (n2-n1+1)  * M.
intros; unfold squared_norm_z, dot_product_z.
apply Rle_trans with (INR (S ((Zabs_nat (n2 - n1)))) * M).
apply squared_norm_z_majoration_aux.
rewrite inj_Zabs_nat; rewrite Zabs_eq; auto with zarith.
rewrite S_INR; rewrite INR_IZR_INZ.
rewrite inj_Zabs_nat; rewrite Zabs_eq; auto with zarith.
rewrite plus_IZR; simpl; right; ring.
Qed.

Lemma Cauchy_Schwarz: forall j1 j2 f1 f2, 
   Rabs (dot_product_z j1 j2 f1 f2)
      <= sqrt (squared_norm_z j1 j2 f1) * sqrt (squared_norm_z j1 j2 f2).
intros.
assert ((2*dot_product_z j1 j2 f1 f2) * (2*dot_product_z j1 j2 f1 f2) - 
  4 * squared_norm_z j1 j2 f1 * squared_norm_z j1 j2 f2 <= 0).
apply discr_neg.
intros.
apply Rle_trans with (squared_norm_z j1 j2 ((scalar_multiply_n x f1)+f2)%type).
apply squared_norm_z_pos.
right.
rewrite squared_norm_n_plus.
rewrite squared_norm_n_scal_mult_n.
rewrite dot_product_z_scal_mult_l.
ring.
apply Rsqr_incr_0_var.
rewrite <- Rsqr_abs.
rewrite Rsqr_mult.
repeat rewrite Rsqr_sqrt; try apply squared_norm_z_pos.
apply Rmult_le_reg_l with (4%R).
fourier.
apply Rplus_le_reg_l with (-4*squared_norm_z j1 j2 f1 * squared_norm_z j1 j2 f2).
ring_simplify.
apply Rle_trans with (2:=H).
unfold Rsqr; right; ring.
apply Rmult_le_pos; apply sqrt_pos.
Qed.


(** Support for Zmin, Zmax in omega *)

Ltac omega_min x y :=
  generalize (Zmin_spec x y); omega.

Ltac omega_max x y :=
  generalize (Zmax_spec x y); omega.

Ltac omega_min_max x y u v :=
  generalize (Zmin_spec x y); 
  generalize (Zmax_spec u v);
  omega.

(* FINITE SUPPORT *)
(** Now, we only consider functions having a finite support (FS) defined by: *)

Definition FS u := exists jinf, exists jsup,  
     forall j, (j < jinf \/ jsup < j)%Z -> u j = 0.

(** Properties on FS *)

Lemma FS_eq: forall f g, FS f -> (forall i, f i = g i) -> FS g.
Proof.
  intros f g (jinf,(jsup,hf)) heq.
  exists jinf; exists jsup; intros.
  rewrite <- heq; auto.
Qed.

Lemma FS_addf : forall f g, FS f -> FS g -> FS (f+g)%type.
Proof.
  unfold FS; intros f g (fmin,(fmax,hf)) (gmin,(gmax,hg)).
  exists (Zmin fmin gmin).
  exists (Zmax fmax gmax).
  intros.
  unfold add_n.
  rewrite hf. rewrite hg.
  ring.
  destruct H. 
  left; omega_min fmin gmin.
  right; omega_max fmax gmax.
  destruct H. 
  left; omega_min fmin gmin.
  right; omega_max fmax gmax.
 Qed.

Lemma FS_add : forall f g, FS f -> FS g -> FS (fun j => f j+g j).
Proof.
  intros; apply FS_addf; trivial.
Qed.

Lemma FS_subf : forall f g, FS f -> FS g -> FS (f-g)%type.
Proof.
  unfold FS; intros f g (fmin,(fmax,hf)) (gmin,(gmax,hg)).
  exists (Zmin fmin gmin).
  exists (Zmax fmax gmax).
  intros; unfold subtract_n.
  rewrite hf. rewrite hg.
  ring.
  destruct H. 
  left; omega_min fmin gmin.
  right; omega_max fmax gmax.
  destruct H. 
  left; omega_min fmin gmin.
  right; omega_max fmax gmax.
 Qed.

Lemma FS_sub : forall f g, FS f -> FS g -> FS (fun j => f j -g j).
Proof.
  intros; apply FS_subf; trivial.
Qed.

Lemma FS_scal_multf : forall l f, FS f -> FS (scalar_multiply_n l f).
Proof.
  intros l f (jinf,(jsup,hf)).
  exists jinf; exists jsup; intros.
  unfold scalar_multiply_n.
  apply Rmult_eq_0_compat_l; auto.
Qed.

Lemma FS_scal_mult : forall l f, FS f -> FS (fun j => l * f j).
Proof.
  intros; apply FS_scal_multf; trivial.
Qed.

Lemma FS_shift1 : forall l f, FS f -> FS (fun j => f (j+l)%Z).
Proof.
  intros l f (jinf,(jsup,hf)).
  exists (jinf-l)%Z; exists (jsup-l)%Z; intros.
  rewrite hf; auto.
  omega.
Qed.

Lemma FS_shift2 : forall l f, FS f -> FS (fun j => f (j-l)%Z).
Proof.
  intros l f hf; apply FS_eq with (fun j => f (j+(-l))%Z); auto.
  apply FS_shift1; auto.
Qed.

Lemma FS_sum_f_R0: forall f n,
  (forall x, (0 <= x <= n)%nat -> FS (f x)) ->
  FS (fun j => sum_f_R0 (fun x => f x j) n).
intros.
induction n; simpl.
apply (H 0%nat); auto.
apply FS_add.
apply IHn.
intros; apply H; destruct H0; split; auto with arith.
apply H; auto with arith.
Qed.

Hint Resolve FS_add FS_addf FS_sub FS_subf FS_scal_mult FS_scal_multf FS_shift1 FS_shift2: FS.

Ltac FS := auto with FS.



(* INFINITE PRODUCT *)

(** We assume there exists an infinite dot product of two functions from Z to R.*)
(**  We also assume that this infinite dot product is equal to the finite dot product previously defined.*)
(**  This only axiom about the dot product allows us to manipulate the dot product of any 2 functions, 
    but all lemmas will require them to have a finite support.  *)

Require Import Epsilon.

(*
Parameter infinite_dot_product : (Z->R) -> (Z->R) -> R.

Axiom infinite_dot_product_eq: forall u v jinf jsup, 
    (forall j, (j < jinf \/ jsup < j)%Z -> u j = 0) ->
    (forall j, (j < jinf \/ jsup < j )%Z-> v j = 0) ->
   infinite_dot_product u v = dot_product_z jinf jsup u v.
*)

(* both u and v are 0 out of interval [jinf,jsup] *)
Definition zero_out u v jinf jsup := 
  (forall j, (j < jinf \/ jsup < j)%Z -> u j = 0) /\
  (forall j, (j < jinf \/ jsup < j )%Z-> v j = 0).

(* dot_product_z gives the same result for 2 intervals, as soon as
   u and v are 0 out of these intervals *)
Lemma dot_product_z_zero_out :
  forall u v a b a' b', zero_out u v a b -> zero_out u v a' b' ->
  dot_product_z a b u v = dot_product_z a' b' u v.
Proof.
  unfold zero_out.
  intros u v a b a' b' hab ha'b'.
  assert (h: (a <= b \/ a > b)%Z) by omega; destruct h.
  assert (h: (a' <= b' \/ a' > b')%Z) by omega; destruct h.
  (* [a,b] and [a',b'] are non-empty intervals *)
  assert (h: (b < a')%Z \/ (a <= a' <= b)%Z \/ (a' <= a <= b')%Z \/ (b' < a)%Z) 
   by omega. destruct h as [h|[h|[h|h]]].
  (* a <= b < a' <= b'  ==> u or v is 0 *)
  assert (forall i, u i = 0 \/ v i = 0).
  intro i; assert (hi: (i < a' \/ b < i)%Z) by omega. destruct hi.
  destruct ha'b' as (h1,h2). left; apply h1. omega.
  destruct hab as (h1,h2). left; apply h1. omega.
  do 2 (rewrite dot_product_z_zero; auto).
  (* a <= a' <= b *)
  assert (hbb': (b <= b' \/ b > b')%Z) by omega; destruct hbb'.
  (* a <= a' <= b <= b' ==> dot_product on [a',b] *)
  transitivity (dot_product_z a' b u v).
  symmetry. apply dot_product_z_zero_left; intuition.
  apply dot_product_z_zero_right; intuition.
  (* a <= a' <= b' <= b ==> dot_product on [a',b'] *)
  transitivity (dot_product_z a b' u v).
  symmetry; apply dot_product_z_zero_right; intuition.
  symmetry; apply dot_product_z_zero_left; intuition.
  (* a' <= a <= b' *)
  assert (hbb': (b <= b' \/ b > b')%Z) by omega; destruct hbb'.
  (* a' <= a <= b <= b' ==> dot_product on [a,b] *)
  transitivity (dot_product_z a' b u v).
  apply dot_product_z_zero_left; intuition.
  apply dot_product_z_zero_right; intuition.
   (* a' <= a <= b' <= b ==> dot_product on [a,b'] *)
  transitivity (dot_product_z a b' u v).
  symmetry. apply dot_product_z_zero_right; intuition.
  apply dot_product_z_zero_left; intuition.

  (* a'<=b' < a <= b  ==> u or v is 0 *)
  assert (forall i, u i = 0 \/ v i = 0).
  intro i; assert (hi: (i < a \/ b' < i)%Z) by omega. destruct hi.
  destruct hab as (h1,h2). left; apply h1. omega.
  destruct ha'b' as (h1,h2). left; apply h1. omega.
  do 2 (rewrite dot_product_z_zero; auto). 
  
  (* either [a,b] or [a',b'] is empty *)
  assert (forall i, u i = 0 \/ v i = 0).
  intro i; destruct ha'b' as (h1,h2).
  left; apply h1.
  omega.
  rewrite dot_product_z_zero; auto.
  rewrite dot_product_z_zero; auto.
  assert (forall i, u i = 0 \/ v i = 0).
  intro i; destruct hab as (h1,h2).
  left; apply h1.
  omega.
  do 2 (rewrite dot_product_z_zero; auto).
Qed.

(* infinite_dot_product is defined using Hilbert's epsilon operator *)

Definition R_inhabited : inhabited R := inhabits 0%R.

Definition infinite_dot_product u v : R :=
  epsilon R_inhabited 
    (fun x => exists jinf, exists jsup,
                    zero_out u v jinf jsup /\ x = dot_product_z jinf jsup u v).

(* but it is actually simpler to use through this lemma *)

Lemma infinite_dot_product_eq: forall u v jinf jsup, 
    (forall j, (j < jinf \/ jsup < j)%Z -> u j = 0) ->
    (forall j, (j < jinf \/ jsup < j )%Z-> v j = 0) ->
   infinite_dot_product u v = dot_product_z jinf jsup u v.
Proof.  
  intros u v a b hu hv.
  assert (hab : zero_out u v a b).
  red; intuition.
  assert (exists a, exists b, zero_out u v a b /\ infinite_dot_product u v = dot_product_z a b u v).
  unfold infinite_dot_product; apply epsilon_spec.
  exists (dot_product_z a b u v); exists a; exists b; intuition.
  destruct H as (a',(b',(h1,h2))).
  rewrite h2; apply dot_product_z_zero_out; auto.
Qed.

Definition infinite_dot_product_dx dx u v := dx * infinite_dot_product u v.

Definition infinite_squared_norm_dx dx u :=  infinite_dot_product_dx dx u u.

Definition infinite_norm_dx dx u :=  sqrt (infinite_squared_norm_dx dx u).

(** Lemmas about the  infinite dot product *)

Lemma infinite_dot_product_FS : forall u v, FS u -> FS v ->
  exists jinf, exists jsup,  infinite_dot_product u v = dot_product_z jinf jsup u v.
Proof.
  intros u v (minu,(maxu,hu)) (minv,(maxv,hv)).
  exists (Zmin minu minv).
  exists (Zmax maxu maxv).
  apply infinite_dot_product_eq.
  intros; apply hu.
  generalize (Zmin_spec minu minv); generalize (Zmax_spec maxu maxv); omega.
  intros; apply hv.
  generalize (Zmin_spec minu minv); generalize (Zmax_spec maxu maxv); omega.
Qed.

Ltac make_it_finite f g :=
  assert (tmp1 : FS f); [FS | 
  assert (tmp2 : FS g); [FS | 
  generalize (infinite_dot_product_FS f g tmp1 tmp2);
  intros (jinf, (jsup, tmp3));
  rewrite tmp3;
  clear tmp1 tmp2 tmp3
 ]].


Lemma infinite_dot_product_dx_zero1:
  forall dx f g, FS g ->
  (forall i, f i = 0) -> 
  infinite_dot_product_dx dx f g = 0.
Proof.
  intros dx f g (jinf,(jsup,hg)) hf.
  unfold infinite_dot_product_dx.
  rewrite (infinite_dot_product_eq f g jinf jsup); trivial.
  unfold dot_product_z.
assert (forall n, forall z,  dx * dot_product_n n z f g=0).
induction n.
intros z; simpl.
rewrite hf; ring.
intros z; simpl.
rewrite Rmult_plus_distr_l.
rewrite IHn.
rewrite hf; ring.
apply H.
Qed.

Lemma infinite_squared_norm_dx_pos: forall dx f, FS f ->
  (0 <= dx) ->
   0 <= infinite_squared_norm_dx dx f.
Proof.
intros dx f (fmin,(fmax,hf)) hdx.
unfold infinite_squared_norm_dx, infinite_dot_product_dx.
rewrite (infinite_dot_product_eq f f fmin fmax); trivial.
fold (squared_norm_z fmin fmax f).
apply Rmult_le_pos; auto.
apply squared_norm_z_pos.
Qed.

Lemma infinite_squared_norm_dx_eq: forall dx f1 f2, FS f1 -> 
  (forall x, f1 x = f2 x) -> 
  infinite_squared_norm_dx dx f1 = infinite_squared_norm_dx dx f2.
Proof.
intros dx f1 f2 (f1min,(f1max,h1)) h.
assert (h2 : forall j : Z, (j < f1min)%Z \/ (f1max < j)%Z -> f2 j = 0).
intros; rewrite <- h; auto.
unfold infinite_squared_norm_dx, infinite_dot_product_dx.
apply f_equal.
rewrite (infinite_dot_product_eq f1 f1 f1min f1max); trivial.
rewrite (infinite_dot_product_eq f2 f2 f1min f1max); trivial.
unfold squared_norm_z, dot_product_z.
assert (forall n, forall z,  dot_product_n n z f1 f1 = dot_product_n n z f2 f2).
  induction n; simpl; intros.
  rewrite h; auto.
  rewrite h; rewrite IHn; auto.
  apply H.
Qed.

Lemma infinite_squared_norm_dx_zero: forall dx f, FS f ->
   (forall x, f x = 0) -> 
  infinite_squared_norm_dx dx f = 0.
Proof.
intros dx f  FSh h.
unfold infinite_squared_norm_dx.
apply infinite_dot_product_dx_zero1; auto.
Qed.

Lemma infinite_squared_norm_dx_homogen: forall dx l f, FS f ->
  infinite_squared_norm_dx dx (scalar_multiply_n l f) = 
    l * l * infinite_squared_norm_dx dx f.
Proof.
  intros.
  unfold infinite_squared_norm_dx, infinite_dot_product_dx.
  destruct H as (jinf,(jsup,H)).
  rewrite infinite_dot_product_eq with (scalar_multiply_n l f) (scalar_multiply_n l f) jinf jsup.
  rewrite infinite_dot_product_eq with f f jinf jsup.
  rewrite dot_product_z_scal_mult_r.
  rewrite dot_product_z_scal_mult_l.
  ring.
  intuition.
  intuition.
  unfold scalar_multiply_n; intuition.
  unfold scalar_multiply_n; intuition.
Qed.  

Lemma infinite_Cauchy_Schwarz_dx: forall dx f1 f2, FS f1 -> FS f2 ->
   0 <= dx ->
   Rabs (infinite_dot_product_dx dx f1 f2)
      <= infinite_norm_dx dx f1 * infinite_norm_dx dx f2.
Proof.
  intros dx f1 f2 h1 h2 hdx.
  unfold infinite_norm_dx, infinite_squared_norm_dx, infinite_dot_product_dx.
  destruct h1 as (jinf1,(jsup1,h1)).
  destruct h2 as (jinf2,(jsup2,h2)).
  pose (min := Zmin jinf1 jinf2).
  pose (max := Zmax jsup1 jsup2).
  rewrite infinite_dot_product_eq with f1 f2 min max.
  rewrite infinite_dot_product_eq with f1 f1 min max.
  rewrite infinite_dot_product_eq with f2 f2 min max.
repeat rewrite  sqrt_mult; trivial; try apply squared_norm_z_pos.
apply Rle_trans with ((sqrt dx * sqrt dx)* 
  (sqrt (dot_product_z min max f1 f1) * 
  sqrt (dot_product_z min max f2 f2))).
2: right; ring.
rewrite Rabs_mult.
apply Rmult_le_compat; try apply Rabs_pos.
rewrite <- sqrt_mult; replace (dx*dx) with dx²;auto with real.
right; apply sym_eq; apply sqrt_Rsqr_abs.
apply Cauchy_Schwarz.
unfold min, max; intros; apply h2; omega_min_max jinf1 jinf2 jsup1 jsup2.
unfold min, max; intros; apply h2; omega_min_max jinf1 jinf2 jsup1 jsup2.
unfold min, max; intros; apply h1; omega_min_max jinf1 jinf2 jsup1 jsup2.
unfold min, max; intros; apply h1; omega_min_max jinf1 jinf2 jsup1 jsup2.
unfold min, max; intros; apply h1; omega_min_max jinf1 jinf2 jsup1 jsup2.
unfold min, max; intros; apply h2; omega_min_max jinf1 jinf2 jsup1 jsup2.
Qed.

Lemma infinite_squared_norm_dx_triang: forall dx f1 f2, FS f1 -> FS f2 ->
  0 <= dx -> 
  infinite_norm_dx dx (fun j => f1 j + f2 j) <= 
      infinite_norm_dx dx f1 + infinite_norm_dx dx f2.
Proof.
  intros dx f1 f2 h1 h2 hdx.
  unfold infinite_norm_dx, infinite_squared_norm_dx, infinite_dot_product_dx.
  destruct h1 as (jinf1,(jsup1,h1)).
  destruct h2 as (jinf2,(jsup2,h2)).
  pose (min := Zmin jinf1 jinf2).
  pose (max := Zmax jsup1 jsup2).
  rewrite infinite_dot_product_eq with (fun j : Z => f1 j + f2 j) (fun j : Z => f1 j + f2 j) min max.
  rewrite infinite_dot_product_eq with f1 f1 min max.
  rewrite infinite_dot_product_eq with f2 f2 min max.
apply Rsqr_incr_0_var.
rewrite Rsqr_sqrt.
unfold Rsqr; rewrite Rmult_plus_distr_l.
rewrite Rmult_plus_distr_r; rewrite Rmult_plus_distr_r.
repeat rewrite sqrt_sqrt.
change (dx * squared_norm_z min max (add_n f1 f2)
  <= dx * squared_norm_z min max f1 +
      sqrt (dx * squared_norm_z min max f2) *
     sqrt (dx * squared_norm_z min max f1) +
     (sqrt (dx * squared_norm_z min max f1) *
     sqrt (dx * squared_norm_z min max f2) +
  dx * squared_norm_z min max f2)).
rewrite squared_norm_n_plus.
apply Rplus_le_reg_l with (-dx*squared_norm_z min max f2
   - dx*squared_norm_z min max f1).
ring_simplify.
repeat rewrite sqrt_mult; trivial.
repeat rewrite Rmult_assoc.
apply Rmult_le_compat_l.
auto with real.
replace (Rmult (sqrt dx)
     (Rmult (sqrt (squared_norm_z min max f2))
        (Rmult (sqrt dx) (sqrt (squared_norm_z min max f1)))))
  with
  ((sqrt dx * sqrt dx) * ((sqrt (squared_norm_z min max f2)) *
     sqrt (squared_norm_z min max f1))) by ring.
rewrite sqrt_sqrt; trivial.
apply Rmult_le_compat_l; trivial.
apply Rle_trans with (1:= RRle_abs (dot_product_z min max f1 f2)).
rewrite Rmult_comm.
apply Cauchy_Schwarz.
apply squared_norm_z_pos.
apply squared_norm_z_pos.
apply Rmult_le_pos; trivial. apply squared_norm_z_pos.
apply Rmult_le_pos; trivial. apply squared_norm_z_pos.
apply Rmult_le_pos; trivial. apply squared_norm_z_pos.
apply Rle_trans with (0+0); auto with real; apply Rplus_le_compat.
apply sqrt_pos.
apply sqrt_pos.
unfold min, max; intros; apply h2; omega_min_max jinf1 jinf2 jsup1 jsup2.
unfold min, max; intros; apply h2; omega_min_max jinf1 jinf2 jsup1 jsup2.
unfold min, max; intros; apply h1; omega_min_max jinf1 jinf2 jsup1 jsup2.
unfold min, max; intros; apply h1; omega_min_max jinf1 jinf2 jsup1 jsup2.
unfold min, max; intros; rewrite h1; [rewrite h2; try ring | idtac];
omega_min_max jinf1 jinf2 jsup1 jsup2.
unfold min, max; intros; rewrite h1; [rewrite h2; try ring | idtac];
omega_min_max jinf1 jinf2 jsup1 jsup2.
Qed.

Lemma infinite_squared_norm_dx_majoration: forall dx f M j1 j2,
   0 <= dx -> 
   (j1 <= j2)%Z ->
   (forall j, (j < j1 \/ j2 < j)%Z -> f j = 0) ->
   (forall j : Z, (j1 <= j <= j2)%Z -> Rsqr (f j) <= M) ->
   infinite_squared_norm_dx dx f <= dx * IZR (j2-j1+1) * M.
Proof.
intros.
unfold infinite_squared_norm_dx, infinite_dot_product_dx.
rewrite (infinite_dot_product_eq f f j1 j2); trivial.
rewrite Rmult_assoc; apply Rmult_le_compat_l; trivial.
apply squared_norm_z_majoration; trivial.
Qed.


Lemma infinite_dot_product_dx_eq_l: forall dx u v w, FS u -> FS w ->
   (forall j, u j = v j) ->
   infinite_dot_product_dx dx u w = infinite_dot_product_dx dx v w.
Proof.
intros dx u v w (umin,(umax,hu)) (wmin,(wmax,hw)) h.
assert (hv : forall j : Z, (j < umin)%Z \/ (umax < j)%Z -> v j = 0).
intros; rewrite <- h; auto.
unfold infinite_dot_product_dx.
apply f_equal.
pose (min := Zmin umin wmin).
pose (max := Zmax umax wmax).
rewrite (infinite_dot_product_eq u w min max); trivial.
rewrite (infinite_dot_product_eq v w min max); trivial.
unfold dot_product_z.
assert (forall n, forall z,  dot_product_n n z u w = dot_product_n n z v w).
  induction n; simpl; intros.
  rewrite h; auto.
  rewrite h; rewrite IHn; auto.
  apply H.
intuition; apply hv;
 unfold min,max in H0; (omega_min umin wmin || omega_max umax wmax).
intuition; apply hw; 
  unfold min,max in H0 ; (omega_min umin wmin || omega_max umax wmax).
intuition; apply hu;
 unfold min,max in H0; (omega_min umin wmin || omega_max umax wmax).
intuition; apply hw; 
  unfold min,max in H0 ; (omega_min umin wmin || omega_max umax wmax).
Qed.

Lemma infinite_dot_product_dx_sym: forall dx u v, FS u -> FS v -> 
   infinite_dot_product_dx dx u v = infinite_dot_product_dx dx v u.
Proof.
intros dx u v (umin,(umax,hu)) (vmin,(vmax,hv)).
unfold infinite_dot_product_dx.
apply f_equal.
pose (min := Zmin umin vmin).
pose (max := Zmax umax vmax).
rewrite (infinite_dot_product_eq u v min max); trivial.
rewrite (infinite_dot_product_eq v u min max); trivial.
apply dot_product_z_sym.
intuition; apply hv;
 unfold min,max in H0; (omega_min umin vmin || omega_max umax vmax).
intuition; apply hu; 
  unfold min,max in H0 ; (omega_min umin vmin || omega_max umax vmax).
intuition; apply hu;
 unfold min,max in H0; (omega_min umin vmin || omega_max umax vmax).
intuition; apply hv; 
  unfold min,max in H0 ; (omega_min umin vmin || omega_max umax vmax).
Qed.

Lemma infinite_dot_product_dx_eq_r: forall dx u v w, FS u ->  FS w ->
   (forall j, u j = v j) ->
   infinite_dot_product_dx dx w u = infinite_dot_product_dx dx w v.
Proof.
  intros.
  rewrite (infinite_dot_product_dx_sym dx w u); auto.
  rewrite (infinite_dot_product_dx_sym dx w v); auto.
  apply infinite_dot_product_dx_eq_l; auto.
  apply FS_eq with u; auto.
Qed.

Ltac omega_min_max3 umin vmin wmin umax vmax wmax :=
  generalize (Zmin_spec vmin wmin);
generalize (Zmin_spec umin (Zmin vmin wmin));
generalize (Zmax_spec vmax wmax);
generalize (Zmax_spec umax (Zmax vmax wmax));
omega.

Lemma infinite_dot_product_dx_add_l: forall dx u v w, FS u -> FS v -> FS w ->
   infinite_dot_product_dx dx (add_n u v) w=
   infinite_dot_product_dx dx u w +  infinite_dot_product_dx dx v w.
Proof.
intros dx u v w (umin,(umax,hu)) (vmin,(vmax,hv)) (wmin,(wmax,hw)).
unfold infinite_dot_product_dx.
replace  (dx * infinite_dot_product u w + dx * infinite_dot_product v w)
with (dx * (infinite_dot_product u w + infinite_dot_product v w)) by ring.
apply f_equal.
pose (min := Zmin umin (Zmin vmin wmin)).
pose (max := Zmax umax (Zmax vmax wmax)).
rewrite (infinite_dot_product_eq (u + v)%type w min max); trivial.
rewrite (infinite_dot_product_eq u w min max); trivial.
rewrite (infinite_dot_product_eq v w min max); trivial.
rewrite <- dot_product_z_distr_l. trivial.
unfold min, max; intros; apply hv; omega_min_max3 umin vmin wmin umax vmax wmax .
unfold min, max; intros; apply hw; omega_min_max3 umin vmin wmin umax vmax wmax .
unfold min, max; intros; apply hu; omega_min_max3 umin vmin wmin umax vmax wmax .
unfold min, max; intros; apply hw; omega_min_max3 umin vmin wmin umax vmax wmax .
unfold min, max; intros; unfold add_n.
rewrite hu. rewrite hv. ring.
 omega_min_max3 umin vmin wmin umax vmax wmax .
 omega_min_max3 umin vmin wmin umax vmax wmax .
unfold min, max; intros; apply hw; omega_min_max3 umin vmin wmin umax vmax wmax .
Qed.

Lemma infinite_dot_product_dx_add_r: forall dx u v w, FS u -> FS v -> FS w ->
   infinite_dot_product_dx dx w (add_n u v) =
   infinite_dot_product_dx dx w u +  infinite_dot_product_dx dx w v.
Proof.
  intros; rewrite infinite_dot_product_dx_sym; FS.
  rewrite infinite_dot_product_dx_add_l; FS.
  rewrite infinite_dot_product_dx_sym; FS.
  apply f_equal.
  rewrite infinite_dot_product_dx_sym; FS.
Qed.

Lemma infinite_dot_product_dx_sub_l: forall dx u v w, FS u -> FS v -> FS w ->
   infinite_dot_product_dx dx (subtract_n u v) w =
   infinite_dot_product_dx dx u w - infinite_dot_product_dx dx v w.
Proof.
intros dx u v w (umin,(umax,hu)) (vmin,(vmax,hv)) (wmin,(wmax,hw)).
unfold infinite_dot_product_dx.
replace  (dx * infinite_dot_product u w - dx * infinite_dot_product v w)
with (dx * (infinite_dot_product u w - infinite_dot_product v w)) by ring.
apply f_equal.
pose (min := Zmin umin (Zmin vmin wmin)).
pose (max := Zmax umax (Zmax vmax wmax)).
rewrite (infinite_dot_product_eq (u - v)%type w min max); trivial.
rewrite (infinite_dot_product_eq u w min max); trivial.
rewrite (infinite_dot_product_eq v w min max); trivial.
rewrite <- dot_product_z_sub_distr_l. trivial.
unfold min, max; intros; apply hv; omega_min_max3 umin vmin wmin umax vmax wmax .
unfold min, max; intros; apply hw; omega_min_max3 umin vmin wmin umax vmax wmax .
unfold min, max; intros; apply hu; omega_min_max3 umin vmin wmin umax vmax wmax .
unfold min, max; intros; apply hw; omega_min_max3 umin vmin wmin umax vmax wmax .
unfold min, max; intros; unfold subtract_n.
rewrite hu. rewrite hv. ring.
 omega_min_max3 umin vmin wmin umax vmax wmax .
 omega_min_max3 umin vmin wmin umax vmax wmax .
unfold min, max; intros; apply hw; omega_min_max3 umin vmin wmin umax vmax wmax .
Qed.

Lemma infinite_dot_product_dx_sub_r: forall dx u v w, FS u -> FS v -> FS w ->
   infinite_dot_product_dx dx w (subtract_n u v) =
   infinite_dot_product_dx dx w u - infinite_dot_product_dx dx w v.
Proof.
  intros; rewrite infinite_dot_product_dx_sym; FS.
  rewrite infinite_dot_product_dx_sub_l; FS.
  rewrite infinite_dot_product_dx_sym; FS.
  apply f_equal.
  rewrite infinite_dot_product_dx_sym; FS.
Qed.

Lemma infinite_dot_product_dx_scal_mult_l :
  forall dx u v l, FS u -> FS v -> 
  infinite_dot_product_dx dx (scalar_multiply_n l u) v =
  l * infinite_dot_product_dx dx u v.
Proof.
intros dx u v l (umin,(umax,hu)) (vmin,(vmax,hv)).
unfold infinite_dot_product_dx.
replace (l * (dx * infinite_dot_product u v)) with (dx * (l * infinite_dot_product u v)) by ring.
apply f_equal.
pose (min := Zmin umin vmin).
pose (max := Zmax umax vmax).
rewrite (infinite_dot_product_eq u v min max); trivial.
rewrite (infinite_dot_product_eq (scalar_multiply_n l u) v min max); trivial.
apply dot_product_z_scal_mult_l.
intuition; unfold scalar_multiply_n; 
rewrite hu; try ring;
 unfold min,max in H0; (omega_min umin vmin || omega_max umax vmax).
intuition; apply hv;
 unfold min,max in H0; (omega_min umin vmin || omega_max umax vmax).
intuition; apply hu; 
  unfold min,max in H0 ; (omega_min umin vmin || omega_max umax vmax).
intuition; apply hv;
 unfold min,max in H0; (omega_min umin vmin || omega_max umax vmax).
Qed.

Lemma infinite_dot_product_dx_scal_mult_r :
  forall dx u v l, FS u -> FS v -> 
  infinite_dot_product_dx dx u (scalar_multiply_n l v) =
  l * infinite_dot_product_dx dx u v.
Proof.
  intros; rewrite infinite_dot_product_dx_sym; FS.
  rewrite (infinite_dot_product_dx_sym dx u v); FS.
  apply infinite_dot_product_dx_scal_mult_l; FS.
Qed.

Lemma infinite_dot_product_dx_shift_indices: forall dx (f1 f2 : Z -> R),
  FS f1 -> FS f2 ->
  infinite_dot_product_dx dx f1 f2 = 
        infinite_dot_product_dx dx (fun j => f1 (j+1)%Z) (fun j => f2 (j+1)%Z).
Proof.
intros dx u v (umin,(umax,hu)) (vmin,(vmax,hv)).
unfold infinite_dot_product_dx.
apply f_equal.
pose (min := Zmin umin vmin).
pose (max := Zmax umax vmax).
pose (l := (Zmin min max - 1)%Z).
pose (r := (Zmax min max + 1)%Z).
rewrite (infinite_dot_product_eq u v l r); trivial.
rewrite (infinite_dot_product_eq (fun j : Z => u (j + 1)%Z) (fun j : Z => v (j + 1)%Z)
  l r); trivial.
rewrite dot_product_z_shift_indices.
rewrite hu.
rewrite hu.
ring.
right; unfold l, r; unfold min, max.
generalize (Zmax_spec (Zmin umin vmin) (Zmax umax vmax)); omega_max umax vmax.
left; unfold l, r; unfold min, max.
generalize (Zmin_spec (Zmin umin vmin) (Zmax umax vmax)); omega_min umin vmin.
unfold l, r. generalize (Zmin_spec min max); omega_max min max.
unfold l, r; unfold min, max; intros; apply hu; intuition.
generalize (Zmin_spec (Zmin umin vmin) (Zmax umax vmax)); omega_min umin vmin.
right; generalize (Zmax_spec (Zmin umin vmin) (Zmax umax vmax)); omega_max umax vmax.
unfold l, r; unfold min, max; intros; apply hv; intuition.
left; generalize (Zmin_spec (Zmin umin vmin) (Zmax umax vmax)); omega_min umin vmin.
right; generalize (Zmax_spec (Zmin umin vmin) (Zmax umax vmax)); omega_max umax vmax.
unfold l, r; unfold min, max; intros; apply hu; intuition.
generalize (Zmin_spec (Zmin umin vmin) (Zmax umax vmax)); omega_min umin vmin.
right; generalize (Zmax_spec (Zmin umin vmin) (Zmax umax vmax)); omega_max umax vmax.
unfold l, r; unfold min, max; intros; apply hv; intuition.
left; generalize (Zmin_spec (Zmin umin vmin) (Zmax umax vmax)); omega_min umin vmin.
right; generalize (Zmax_spec (Zmin umin vmin) (Zmax umax vmax)); omega_max umax vmax.
Qed.

(* FS *)

(* Lemme sur Ah et infinite_dot_product_dx *)
Lemma semi_infinite_dot_product_dx: forall dx Ah u v,
  (forall u, FS u -> FS (Ah u)) ->
  (forall u v, FS u -> FS v ->
    infinite_dot_product_dx dx (Ah u) v
   = infinite_dot_product_dx dx u (Ah v)) ->
  FS u -> FS v ->
  infinite_dot_product_dx dx (Ah u) v =
    /4 * infinite_dot_product_dx dx (Ah (add_n u v)) (add_n u v)
   - /4 * infinite_dot_product_dx dx (Ah (subtract_n u v)) (subtract_n u v).
intros.
rewrite infinite_dot_product_dx_sub_r; FS.
rewrite infinite_dot_product_dx_add_r; FS.
repeat rewrite H0; FS.
repeat rewrite infinite_dot_product_dx_sub_l; FS.
repeat rewrite infinite_dot_product_dx_add_l; FS.
repeat rewrite <- (H0 u); FS.
repeat rewrite <- infinite_dot_product_dx_sym with (v:= Ah u); FS.
field.
Qed.


Lemma infinite_squared_norm_parall: forall dx x y, FS x -> FS y ->
 infinite_squared_norm_dx dx (x+y)%type +
  infinite_squared_norm_dx dx (x-y)%type =
   2*(infinite_squared_norm_dx dx x + infinite_squared_norm_dx dx y).
intros.
unfold infinite_squared_norm_dx.
rewrite infinite_dot_product_dx_add_r; FS.
repeat rewrite infinite_dot_product_dx_add_l; FS.
rewrite infinite_dot_product_dx_sub_r; FS.
repeat rewrite infinite_dot_product_dx_sub_l; FS.
ring.
Qed.

Lemma infinite_squared_norm_parall_add: forall dx x y, 
   0 <= dx ->
   FS x -> FS y ->
   infinite_squared_norm_dx dx (x+y)%type <= 
  2*(infinite_squared_norm_dx dx x + infinite_squared_norm_dx dx y).
intros.
rewrite <- infinite_squared_norm_parall; trivial.
rewrite <- (Rplus_0_r (infinite_squared_norm_dx dx (x + y)%type)) at 1.
apply Rplus_le_compat_l.
apply infinite_squared_norm_dx_pos; FS.
Qed.

Lemma infinite_squared_norm_parall_sub: forall dx x y, 
   0 <= dx -> FS x -> FS y ->
   infinite_squared_norm_dx dx (x-y)%type <= 
  2*(infinite_squared_norm_dx dx x + infinite_squared_norm_dx dx y).
intros.
rewrite <- infinite_squared_norm_parall; trivial.
rewrite <- (Rplus_0_l (infinite_squared_norm_dx dx (x - y)%type)) at 1.
apply Rplus_le_compat_r.
apply infinite_squared_norm_dx_pos; FS.
Qed.


