(** R^2, with basic operations and norm *)

Require Export Reals.
Require Export Reals_compl.
Open Local Scope nat_scope.
Open Local Scope R_scope.

(** R^2 is a vector space. *)

Definition add2 (X1 X2 : R * R) : R * R :=
  (fst X1 + fst X2, snd X1 + snd X2).

Definition subtract2 (X1 X2 : R * R) : R * R :=
  (fst X1 - fst X2, snd X1 - snd X2).

Definition scalar_multiply2 (l : R) (X : R * R) : R * R :=
  (l * fst X, l * snd X).

Notation "f + g" := (add2 f g) : type_scope.
Notation "f - g" := (subtract2 f g) : type_scope.


(** R^2 is an Euclidean space. *)

Definition norm_l2 (X : R * R) : R :=
  sqrt (Rsqr (fst X) + Rsqr (snd X)).

Lemma norm_l2_pos :
  forall X : R * R,
  0 <= norm_l2 X.
intros; unfold norm_l2.
apply sqrt_pos.
Qed.

Lemma norm_l2_scalar_mult: forall X l,
  norm_l2 (scalar_multiply2 l X) = Rabs l*norm_l2 X.
intros; unfold norm_l2, scalar_multiply2; simpl.
replace (Rsqr (l * fst X) + Rsqr (l * snd X)) with ((Rsqr l)*(Rsqr (fst X) + Rsqr (snd X))).
rewrite sqrt_mult; auto with real.
now rewrite sqrt_Rsqr_abs.
apply Rle_trans with (0+0); auto with real.
unfold Rsqr; ring.
Qed.

Lemma norm_l2_pos_strict :
  forall X : R * R,
  ~(X = (0,0)) -> 0 < norm_l2 X.
intros; unfold norm_l2.
assert (0 < Rsqr (fst X) + Rsqr (snd X)).
case (Req_dec (fst X) 0); intros.
rewrite H0; apply Rlt_le_trans with (Rsqr (snd X)).
apply Rlt_0_sqr.
unfold not; intros; apply H; destruct X.
simpl in H0, H1; rewrite H0; rewrite H1; auto with real.
unfold Rsqr; right; ring.
apply Rle_lt_trans with (0+Rsqr (snd X))%R.
ring_simplify; auto with real.
apply Rplus_lt_compat_r; apply Rlt_0_sqr; auto.
rewrite <- sqrt_0; apply sqrt_lt_1; auto with real.
Qed.

Lemma norm_l2_proj1 :
  forall x : R,
  norm_l2 (x,0) = Rabs x.
intros; unfold norm_l2; simpl.
replace  (Rsqr x + Rsqr 0)%R with (Rsqr x)%R by (unfold Rsqr; ring).
apply sqrt_Rsqr_abs.
Qed.

Lemma norm_l2_proj2 :
  forall y : R,
  norm_l2 (0,y) = Rabs y.
intros; unfold norm_l2; simpl.
replace  (Rsqr 0 + Rsqr y)%R with (Rsqr y)%R by (unfold Rsqr; ring).
apply sqrt_Rsqr_abs.
Qed.

Lemma Rabs_fst_le_norm_l2: forall X,
  Rabs (fst X) <= norm_l2 X.
unfold norm_l2; intros.
rewrite <- sqrt_Rsqr_abs.
apply sqrt_le_1; auto with real.
apply Rle_trans with (0+0); auto with real.
apply Rle_trans with (Rsqr (fst X)+0); auto with real.
Qed.

Lemma Rabs_snd_le_norm_l2: forall X,
  Rabs (snd X) <= norm_l2 X.
unfold norm_l2; intros.
rewrite <- sqrt_Rsqr_abs.
apply sqrt_le_1; auto with real.
apply Rle_trans with (0+0); auto with real.
apply Rle_trans with (0 + Rsqr (snd X)); auto with real.
Qed.


(** Functions from R^2 to R. *)
Section defs.
Variable A B:Type.

Definition add2_fun (f g : R * R -> R) : R * R -> R :=
  fun X : R * R => f X + g X.

Definition subtract2_fun (f g : R * R -> R) : R * R -> R :=
  fun X : R * R => f X - g X.

Definition scalar_multiplyA_fun (l : R) (f : A -> R) : A -> R :=
  fun X : A => l * f X.

Definition multiply2_fun (f g : R * R -> R) : R * R -> R :=
  fun X : R * R => f X * g X.

Definition divide2_fun (f g : R * R -> R) : R * R -> R :=
  fun X : R * R => f X / g X.

Definition scalar_multiplyAA_fun (l:R) (f:A->B->R) :A->B->R :=
  fun X Y=> l*f X Y. 

End defs.