(** Convergence of the numerical scheme *)

Require Export stability.
Require Export consistency.
Open Local Scope nat_scope.
Open Local Scope R_scope.

Section Conv.

Variable c:R.
Variable u0 u1: R->R.
Variable f : R*R -> R.
Variable usol : R * R -> R.
Variable u0h u1h: Z-> R.
Variable fh: Z -> nat -> R.
Variable unh : R * R -> Z -> nat -> R.
Variable eps eta A B:R.

(** All the hypotheses *)

Hypothesis usol_is_sol: is_solution c u0 u1 f usol.
Hypothesis unh_is_discrete_sol: forall dX, is_discrete_solution c u0h u1h fh unh dX.

Hypothesis u0_zero: forall x, (x < A)  \/ (B < x)  -> u0 x = 0.
Hypothesis u1_zero: forall x, (x < A)  \/ (B < x)  -> u1 x = 0.
Hypothesis f_zero: forall x y, (x < A - c*y)  \/ (B+c*y < x)  -> f (x,y) = 0.

Hypothesis u0h_is_u0: forall dX j, u0h j = u0 (fst (Xjk dX j 0)).
Hypothesis u1h_is_u1: forall dX j, u1h j = u1 (fst (Xjk dX j 1)).
Hypothesis fh_is_f: forall dX j k, fh j k = f (Xjk dX j k).

Hypothesis f_is_zero_at_beginning: forall x, f (x,0) = 0.
Hypothesis c_pos: 0 < c.
Hypothesis eps_lt: 0 < eps < 1.
Hypothesis T_pos: 0 < T.
Hypothesis A_le_B: A <= B.
Hypothesis eta_pos: 0 < eta.

Notation OupsCFL_l := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX /\ CFL_condition_l c eta dX)).
Notation OupsCFL := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX /\ CFL_condition_r c eps dX /\ CFL_condition_l c eta dX)).

(** A few finite support lemmas *)

Lemma FS_convergence_error_aux: forall dX k, (0 < fst dX) -> (0 < snd dX) ->
 (forall j, (j < Int_part (A/fst dX) - Z_of_nat k * up (c * snd dX / fst dX)
            \/ Int_part (B/fst dX) + Z_of_nat k * up (c * snd dX / fst dX) < j)%Z -> 
   convergence_error usol unh dX j k = 0).
intros dX k L1 L1' j H.
unfold convergence_error, ubar, discretization.
rewrite (zeroness_unh c u0 u1 f u0h u1h fh unh dX A B); trivial.
rewrite (zeroness_usol c u0 u1 f usol A B); trivial.
ring.
apply zeroness_unh_aux2; trivial.
Qed.


Lemma FS_convergence_error: forall dX k, (0 < fst dX) -> (0 < snd dX) ->
 FS (fun j => convergence_error usol unh dX j k).
intros dX k L1 L2.
exists (Int_part (A/fst dX) - Z_of_nat k* up (c * snd dX / fst dX))%Z; 
   exists  (Int_part (B/fst dX) + Z_of_nat k* up (c * snd dX / fst dX))%Z.
apply FS_convergence_error_aux; trivial.
Qed.

Hint Resolve FS_convergence_error: FS.


Lemma unh_0: forall dX j, unh dX j 0 = ubar usol dX j 0.
intros.
generalize  (unh_is_discrete_sol dX j 0%nat).
unfold scheme,Lh.
replace (u0h j) with (ubar usol dX j 0);auto with real.
rewrite u0h_is_u0 with dX j.
elim usol_is_sol;unfold L0.
intros H1 (H2,H3).
unfold ubar,discretization,Xjk.
rewrite Rmult_0_l;apply H3.
Save.

(** The convergence error is solution of the same scheme with specific inputs *)

Lemma convergence_error_is_discrete_sol: forall dX j n, 
  snd dX <>0 ->
  scheme c
      (fun j=> 0) 
      (fun j => (convergence_error usol unh dX j 1) / (snd dX)) 
      (fun j m => truncation_error c usol u0h u1h fh dX j (S m))
      (convergence_error usol unh) dX j n = 0.
unfold is_discrete_solution, scheme in unh_is_discrete_sol.
unfold scheme, convergence_error; intros.
case n.
rewrite Lh_is_linear.
apply trans_eq with ((Lh c (ubar usol dX) dX j 0 - u0h j) -
   (Lh c (unh dX) dX j 0 - u0h j));[ring|idtac].
rewrite (unh_is_discrete_sol dX j 0%nat).
rewrite u0h_is_u0 with dX j.
simpl.
unfold ubar.
unfold discretization.
elim usol_is_sol.
unfold L0;intros H1 (H2,H3).
 unfold Xjk.
simpl.
rewrite Rmult_0_l.
rewrite H3.
ring.
intros n0; case n0.
rewrite Lh_is_linear.
simpl.
rewrite unh_0.
rewrite Ah_eq with (1:=unh_0 dX).
field;auto.
intros.
rewrite Lh_is_linear.
apply trans_eq with ((-(Lh c (unh dX) dX j (S (S n1)) - fh j (S n1)) +
   (-fh j (S n1)+ Lh c (ubar usol dX) dX j (S (S n1)) 
   - truncation_error c usol u0h u1h fh dX j (S (S n1)))));[ring|idtac].
rewrite (unh_is_discrete_sol dX j (S (S n1))).
unfold truncation_error, scheme.
ring.
Qed.

(** Initializations are ok *)

Lemma condinite1j_aux: forall ph qh dX,
  (snd dX <> 0) ->
  (forall j, Lh c ph dX j 1 = Lh c qh dX j 1) ->
  (forall j, ph j 0%nat = qh j 0%nat) ->
  (forall j, ph j 1%nat = qh j 1%nat).
clear.
unfold Lh; intros.
apply Rmult_eq_reg_l with (/snd dX); auto with real.
apply Rplus_eq_reg_l with  ((-ph j 0%nat) / snd dX +
     snd dX / 2 * Ah c (fun j' : Z => ph j' 0%nat) (fst dX) j).
apply trans_eq with ((ph j 1%nat - ph j 0%nat) / snd dX +
     snd dX / 2 * Ah c (fun j' : Z => ph j' 0%nat) (fst dX) j).
unfold Rdiv; ring.
rewrite H0.
unfold Ah; repeat rewrite H1.
unfold Rdiv; ring.
Qed.

Lemma e1h_LL1h: forall x dX, fst dX<>0 -> snd dX<>0->
  convergence_error usol unh dX (jx x (fst dX)) 1 / (snd dX) =
  LL1h c usol (Xxt (x,0) dX) dX. 
intros; unfold convergence_error,LL1h.
elim usol_is_sol.
unfold L1.
intros H1 (H2,H3).
unfold Xxt. 
simpl.
replace (kt 0 (snd dX)) with 0%nat.
unfold Xjk.
rewrite Rmult_0_l.
rewrite H2.
set (j := jx x (fst dX)).
replace (u1 (IZR j * fst dX)) with 
   (Lh c (unh dX) dX j 1).
unfold Lh,Ah.
repeat rewrite unh_0.
unfold ubar,discretization.
unfold Xjk.
rewrite Rmult_0_l.
replace (IZR j * fst dX + fst dX) with (IZR (j + 1) * fst dX).
replace (IZR j * fst dX - fst dX) with (IZR (j - 1) * fst dX).
rewrite Rmult_1_l,Rplus_0_l.
field.
now split.
unfold Zminus.
rewrite plus_IZR, Ropp_Ropp_IZR.
simpl;ring.
rewrite plus_IZR;simpl;ring. 
generalize (unh_is_discrete_sol dX j 1%nat).
unfold scheme.
replace (u1h j) with (u1 (IZR j * fst dX));auto with real.
rewrite (u1h_is_u1 dX).
trivial.
unfold kt.
replace (0 / snd dX) with (IZR 0).
now rewrite Int_part_IZR.
unfold Rdiv.
now rewrite Rmult_0_l.
Qed.


(* lemme 72-210709*)
Lemma condinite1j: 
  is_sufficiently_regular_n 4%nat usol ->
  Oups 
   (fun x dX => convergence_error usol unh dX (jx x (fst dX)) 1 / (snd dX))
   (fun dX => fst dX ^2 + snd dX^2).
intros.
apply OuP_eq_trans_l with (fun x dX => LL1h c usol (Xxt (x,0) dX) dX).
intros.
rewrite e1h_LL1h;trivial.
elim H0;auto with real.
elim H0;auto with real.
destruct LL1h_is_second_order with (1:=usol_is_sol) (2:=f_is_zero_at_beginning) (3:=H) 
  as (C ,(alpha,(H2,(H3,H4)))).
exists C;exists alpha.
intuition.
unfold Xxt,Xjk,kt.
simpl.
replace (0 / snd dX) with (IZR 0).
rewrite Int_part_IZR.
simpl.
rewrite Rmult_0_l.
apply H4;auto.
unfold Rdiv;simpl;ring.
Qed. 

Lemma condinite1h_aux: forall (P:R*R-> Prop) h g j1 j2 C1 C2,
   (forall dX, P dX -> 0 < fst dX) ->
   0 < C1 -> 0 <= C2 ->
  (forall dX, P dX -> (j1 dX <= j2 dX)%Z) ->
  (forall dX, P dX -> (fst dX) *IZR (j2 dX - j1 dX+1) <= C1+C2*norm_l2 dX) ->
  (forall dX j, P dX -> (j < j1 dX)%Z \/ (j2 dX < j)%Z -> h dX j = 0) ->
  OuP P (fun (x:R) dX => Rsqr (h dX (jx x (fst dX)))) g -> 
  OuP P (fun (_:unit) dX => infinite_squared_norm_dx (fst dX) (h dX)) g.
unfold OuP; intros P h g j1 j2 C1 C2 Y1 L LL L0 L1 L2 (alp,(C,(H1,(H2,H3)))).
exists alp; exists (C*(C1+C2*alp)).
split; trivial.
split.
apply Rmult_lt_0_compat; auto with real.
apply Rlt_le_trans with (C1+0); auto with real.
ring_simplify; exact L.
apply Rplus_le_compat_l; apply Rmult_le_pos; auto with real.
intros.
apply Rle_trans with (((fst dX)* IZR (j2 dX-j1 dX+1))*(C*Rabs (g dX))).
rewrite Rabs_right.
apply infinite_squared_norm_dx_majoration; auto with real.
intros.
rewrite Rsqr_abs.
replace j with (jx (IZR j*fst dX) (fst dX)).
unfold Rsqr; rewrite <- Rabs_mult.
apply H3; auto.
unfold jx.
replace (IZR j * fst dX / fst dX) with (IZR j).
apply Int_part_IZR.
field.
apply Rgt_not_eq; apply Y1; trivial.
apply Rle_ge; apply infinite_squared_norm_dx_pos.
exists (j1 dX); exists (j2 dX); auto.
left; apply Y1; trivial.
apply Rle_trans with ((C1+C2*alp)*(C * Rabs (g dX)));[idtac|right; ring].
apply Rmult_le_compat_r; auto with real.
apply Rmult_le_pos; auto with real.
apply Rabs_pos.
apply Rle_trans with (1:=L1 dX H0).
apply Rplus_le_compat_l.
apply Rmult_le_compat_l; auto with real.
Qed.


(* lemme 73-210709 *)
Lemma condinite1h: 
 is_sufficiently_regular_n 4%nat usol ->
  Oups
   (fun (_:unit) dX => infinite_norm_dx (fst dX)
               (fun j => convergence_error usol unh dX j 1 / (snd dX)))
   (fun dX => fst dX ^2 +snd dX^2).
intros.
apply OuP_sqrt.
intros _ dX (T1,T2).
apply infinite_squared_norm_dx_pos.
exists (Int_part (A/fst dX) - Z_of_nat 1* up (c * snd dX / fst dX))%Z; 
   exists  (Int_part (B/fst dX) + Z_of_nat 1* up (c * snd dX / fst dX))%Z.
intros; rewrite FS_convergence_error_aux; trivial.
unfold Rdiv; ring.
now left.
apply condinite1h_aux with  (fun dX => (Int_part (A/fst dX) - Z_of_nat 1* up (c * snd dX / fst dX))%Z) 
            (fun dX => (Int_part (B/fst dX) + Z_of_nat 1* up (c * snd dX / fst dX))%Z) (B-A+1) (2*c+4)%R.
intros dX (Y1,_); assumption.
apply Rplus_lt_reg_r with A; ring_simplify.
apply Rle_lt_trans with (1:=A_le_B); auto with real.
fourier.
intros dX (T1,T2); (simpl (Z_of_nat 1)).
assert (0 < up (c * snd dX / fst dX))%Z.
apply up_gt_zero.
repeat apply Rmult_lt_0_compat; auto with real.
apply Zle_trans with (Int_part (A / fst dX)); auto with zarith.
apply Zle_trans with (Int_part (B / fst dX)); auto with zarith.
apply Int_part_monotone.
unfold Rdiv; apply Rmult_le_compat_r; auto with real.
intros dX (T1,T2); simpl (Z_of_nat 1).
unfold Zminus; repeat rewrite plus_IZR.
rewrite Ropp_Ropp_IZR; rewrite plus_IZR; rewrite mult_IZR.
rewrite Ropp_Ropp_IZR;rewrite mult_IZR; simpl (IZR 1).
apply Rle_trans with
  ((fst dX *(IZR (Int_part (B / fst dX)))
    + fst dX * (-1 + - (IZR (Int_part (A / fst dX))))
    +2* fst dX *IZR (up (c * snd dX / fst dX))
    +2*fst dX)).
right; ring.
unfold Rminus; repeat rewrite Rplus_assoc.
apply Rplus_le_compat.
apply Rle_trans with (fst dX * (B / fst dX)).
apply Rmult_le_compat_l; auto with real.
eapply base_Int_part.
right; field; auto with real.
apply Rplus_le_compat.
apply Rmult_le_reg_l with (/fst dX); auto with real.
apply Rplus_le_reg_l with (IZR (Int_part (A / fst dX))).
apply Rle_trans with (-1);[right; field; auto with real|idtac].
apply Rle_trans with (IZR (Int_part (A/fst dX)) - A/fst dX);[idtac|right; field; auto with real].
left; eapply base_Int_part.
apply Rle_trans with (2*fst dX * (c * snd dX / fst dX +1) + 2 * fst dX).
apply Rplus_le_compat_r.
apply Rmult_le_compat_l.
apply Rmult_le_pos; auto with real.
apply Rplus_le_reg_l with (-( c * snd dX / fst dX)); ring_simplify.
rewrite Rplus_comm; eapply archimed.
apply Rle_trans with (2*c*snd dX +4*fst dX).
right; field; auto with real.
apply Rle_trans with  ((2*c+4) * norm_l2 dX).
rewrite Rmult_plus_distr_r with (r3:=norm_l2 dX).
apply Rplus_le_compat; apply Rmult_le_compat_l.
apply Rmult_le_pos; auto with real.
apply Rle_trans with (2:=Rabs_snd_le_norm_l2 _).
rewrite Rabs_right; auto with real.
left; auto with real.
apply Rmult_le_pos; auto with real.
apply Rle_trans with (2:=Rabs_fst_le_norm_l2 _).
rewrite Rabs_right; auto with real.
left; auto with real.
apply Rle_trans with  (0+(2*c+4) * norm_l2 dX); auto with real.
intros  dX j (T1,T2) T4.
rewrite FS_convergence_error_aux; trivial.
unfold Rdiv; ring.
apply OuP_Rsqr.
intros; apply Rle_0_sqr.
apply OuP_eq_trans_r with  (fun dX : R * R => (fst dX ^ 2 + snd dX ^ 2)).
intros.
rewrite sqrt_Rsqr; auto with real.
apply Rle_trans with (0+0); try apply Rplus_le_compat;auto with real.
unfold pow; simpl; rewrite Rmult_1_r; fold (Rsqr (fst dX)); auto with real.
unfold pow; simpl; rewrite Rmult_1_r; fold (Rsqr (snd dX)); auto with real.
apply condinite1j; auto.
Qed.

(*lemme 74-210709*)
Lemma energyiniteh:
 is_sufficiently_regular_n 4%nat usol ->
 Oups
  (fun (_:unit) dX => 
    sqrt (energy c (convergence_error usol unh dX) dX 0))
  (fun dX => fst dX ^2 +snd dX^2).
intros.
unfold energy.
apply OuP_eq_trans_l with
  (fun (_ : unit) (dX : R * R) =>
   sqrt
     (1 / 2 *
      infinite_squared_norm_dx (fst dX)
        (scalar_multiply_n (1 / snd dX)
           ((fun j : Z => convergence_error usol unh dX j 1))))).
intros.
rewrite infinite_dot_product_dx_zero1.
apply f_equal.
ring_simplify.
apply f_equal.
apply infinite_squared_norm_dx_eq.
destruct H0 as (H0,H1); FS.
intros.
unfold scalar_multiply_n, subtract_n.
simpl.
replace (convergence_error usol unh dX x 0) with 0.
ring.
unfold convergence_error; simpl.
rewrite unh_0; ring.
destruct H0 as (H0,H1); FS.
intros.
unfold Ah.
unfold convergence_error; simpl.
repeat rewrite unh_0.
unfold Rdiv; ring.
apply OuP_eq_trans_l with
  (scalar_multiplyAA_fun _  _  (sqrt (/2))
    (fun (_ : unit) (dX : R * R) =>
            (infinite_norm_dx  (fst dX)
               (fun j : Z => convergence_error usol unh dX j 1 / snd dX)))).
intros.
unfold scalar_multiplyAA_fun, scalar_multiply_n.
unfold infinite_norm_dx; rewrite <- sqrt_mult.
apply f_equal.
replace (/2) with (1/2) by field.
apply f_equal.
apply infinite_squared_norm_dx_eq; trivial.
destruct H0 as (H0,H1).
exists (Int_part (A/fst dX) - Z_of_nat 1* up (c * snd dX / fst dX))%Z; 
   exists  (Int_part (B/fst dX) + Z_of_nat 1* up (c * snd dX / fst dX))%Z.
intros; rewrite FS_convergence_error_aux; trivial.
unfold Rdiv; ring.
intros; unfold Rdiv; ring.
auto with real.
apply infinite_squared_norm_dx_pos.
destruct H0 as (H0,H1).
exists (Int_part (A/fst dX) - Z_of_nat 1* up (c * snd dX / fst dX))%Z;
   exists  (Int_part (B/fst dX) + Z_of_nat 1* up (c * snd dX / fst dX))%Z.
intros; rewrite FS_convergence_error_aux; trivial.
unfold Rdiv; ring.
left; now destruct H0.
apply OuP_multiplyAA.
apply condinite1h; trivial.
Qed.


(** The sum of all truncation errors is small *)

(* consistance lemme 75-210709*)
Lemma sum_truncation_error:  
   is_sufficiently_regular_n 4 usol  -> 
   OupsCFL_l  
  (fun (t:{t|Under T t}) dX =>
    snd dX * sum_f 1 (kt (projT1 t) (snd dX)-1)
    (fun n => (infinite_norm_dx (fst dX)
                   (fun j => truncation_error c usol u0h u1h fh dX j  (S n)))))
  (fun dX => (fst dX)^2+ (snd dX)^2).
intros.
apply OuP_sum_f_minus1_bis with
 (f:=fun dX n => 
        (infinite_norm_dx (fst dX)
           (fun j : Z => truncation_error c usol u0h u1h fh dX j n))).
apply T_pos.
intros dX (T1,(T2,T3)); trivial.
assert (scheme_is_consistent_of_order c usol u0h u1h fh eta 2%nat 2%nat).
apply consistency with u0 u1 f A B; auto.
unfold scheme_is_consistent_of_order in H0.
apply OuP_sqrt; trivial.
intros.
apply infinite_squared_norm_dx_pos.
apply FS_truncation_error with u0 u1 f A B; trivial; try eapply H1.
left; apply H1.
Qed.


(** The energy of the convergence error is small *)

Lemma energie: (*81p44- 140909*)
   is_sufficiently_regular_n 4 usol  -> 
 OupsCFL 
  (fun (t:{t|Under T t}) dX => 
    sqrt (energy c (convergence_error usol unh dX) dX
                            (kt (projT1 t) (snd dX) -1)%nat))
  (fun dX => fst dX ^2 +snd dX^2).
intros J2.
apply OuP_maj with 
  (fun (t : {t : R | Under T t}) => add2_fun
     (fun (dX : R * R) =>
        sqrt (energy c (convergence_error usol unh dX) dX  0))
     (scalar_multiplyA_fun _ (sqrt 2 / (2*sqrt (2*eps-Rsqr eps)))
       (fun dX => (snd dX) 
             * sum_f 1  (kt (projT1 t) (snd dX) - 1)
                     (fun n => (infinite_norm_dx (fst dX)
                               (fun j => truncation_error c usol u0h u1h fh dX j (S n))))))).
intros X dX (Y1,(Y2,(Y3,Y4))).
unfold add2_fun, scalar_multiplyA_fun.
repeat rewrite Rabs_right.
rewrite <- Rmult_assoc.
apply energy_majoration with (fun _ : Z => 0)  
  (fun j0 : Z => convergence_error usol unh dX j0 1 / snd dX) 
  (up (c * snd dX / fst dX)) 
  (Int_part (A/fst dX) - Z_of_nat 1* (up (c * snd dX / fst dX)))%Z  
  (Int_part (B/fst dX) + Z_of_nat 1* (up (c * snd dX / fst dX)))%Z; 
   trivial.
unfold is_discrete_solution; intros.
apply convergence_error_is_discrete_sol.
apply Rgt_not_eq; apply Y2.
apply up_gt_zero.
repeat apply Rmult_lt_0_compat; auto with real.
intros.
rewrite FS_convergence_error_aux; trivial; try apply H.
unfold Rdiv; ring.
intros.
apply FS_truncation_error_aux with u0 u1 f A B; trivial.
replace  (Z_of_nat (S k)) with (Z_of_nat 1 +Z_of_nat k)%Z.
destruct H;[left|right].
apply Zlt_le_trans with (1:=H);apply Zeq_le; ring.
apply Zle_lt_trans with (2:=H);apply Zeq_le; ring.
rewrite (inj_S k); simpl (Z_of_nat 1); unfold Zsucc; ring.
apply Rle_ge.
replace 0 with (0+0) by ring.
apply Rplus_le_compat.
apply sqrt_pos.
repeat apply Rmult_le_pos.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
apply Rmult_lt_0_compat.
auto with real.
apply sqrt_lt_R0.
replace (2*eps - Rsqr eps) with (eps * (2-eps)) by (unfold Rsqr; ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps; ring_simplify.
apply Rlt_trans with 1.
apply eps_lt.
auto with real.
left; now apply Y2.
apply sum_f_pos.
intros; apply sqrt_pos.
apply Rle_ge; apply sqrt_pos.
apply OuP_is_EV_plus.
apply OuP_unit.
apply OuP_implies_OuP with 
  (P1:=(fun dX => 0 < fst dX /\ 0 < snd dX)).
clear; intuition.
apply energyiniteh; trivial.
apply OuP_multiplyAA.
apply OuP_implies_OuP with (2:=sum_truncation_error J2). 
intros dX (H1,(H2,(H3,H4))); split; try split; trivial.
Qed.


(** The scheme is convergent *)

Lemma convergence: 
  is_sufficiently_regular_n 4%nat usol  -> 
  scheme_is_convergent_of_order c usol unh eps eta 2%nat 2%nat.
intros.
unfold scheme_is_convergent_of_order.
apply OuP_maj with
  (fun (t : {t : R | Under T t}) (dX : R * R) =>
    (sqrt 2 / (sqrt (2*eps-Rsqr eps))) * (snd dX) *
    sum_f 1 (kt (projT1 t) (snd dX)) 
              (fun n => sqrt (energy c (convergence_error usol unh dX) dX (n-1)))).
                                       (* was unh *)
intros X dX (H1,(H2,(H3,H4))).
rewrite Rabs_right.
2: apply Rle_ge; apply sqrt_pos.
case_eq (kt (projT1 X) (snd dX)).
intros G.
unfold sum_f; simpl.
unfold infinite_norm_dx; rewrite infinite_squared_norm_dx_zero; FS.
rewrite sqrt_0.
apply Rabs_pos.
intros; unfold convergence_error.
rewrite unh_0; ring.
intros n G.
apply Rle_trans with 
 (sum_f 1 (S n)
     (fun n : nat =>  (infinite_norm_dx (fst dX)
         (fun j => convergence_error usol unh dX j n
                           - convergence_error usol unh dX j (n-1))))).
apply Rle_trans with
  ((infinite_norm_dx (fst dX)
     (fun j => (sum_f 1 (S n)
     (fun n : nat  => convergence_error usol unh dX j n
                           - convergence_error usol unh dX j (n-1)))))).
right.
unfold infinite_norm_dx; apply f_equal.
apply infinite_squared_norm_dx_eq; FS; auto with real.
intros.
clear G; induction n.
unfold sum_f; simpl.
unfold convergence_error at 3.
rewrite unh_0.
ring.
unfold sum_f; simpl.
unfold sum_f in IHn; simpl in IHn.
replace (n-0)%nat with n in IHn by auto with arith.
rewrite <- IHn.
replace (n+1-0)%nat with (n+1)%nat by auto with arith.
replace (S n) with (n+1)%nat.
ring.
omega.
clear G; induction n.
unfold sum_f; simpl.
apply Rle_refl.
unfold sum_f in *; simpl in *.
eapply Rle_trans.
apply infinite_squared_norm_dx_triang; FS.
apply FS_sum_f_R0; FS.
auto with real.
apply Rplus_le_compat_r.
replace (n-0)%nat with n in IHn by auto with arith.
apply IHn.
rewrite Rabs_right.
clear G; induction n.
unfold sum_f; simpl.
change 1%nat with (0+1)%nat.
apply minorationenergy2 with (fun _ => 0) 
   (fun j0 : Z => convergence_error usol unh dX j0 1 / snd dX)
    (fun (j0 : Z) (m : nat) =>
          truncation_error c usol u0h u1h fh dX j0 (S m))
    (up (c * snd dX / fst dX))
    (Int_part (A/fst dX)-(up (c * snd dX / fst dX)))%Z (Int_part (B/fst dX)+(up (c * snd dX / fst dX)))%Z; trivial.
unfold is_discrete_solution; intros.
apply convergence_error_is_discrete_sol; auto with real.
apply up_gt_zero.
repeat apply Rmult_lt_0_compat; auto with real.
intros; rewrite FS_convergence_error_aux; trivial.
unfold Rdiv; ring.
replace (Z_of_nat 1 * up (c * snd dX / fst dX))%Z with (up (c * snd dX / fst dX)); try exact H0.
simpl (Z_of_nat 1); ring.
intros; apply FS_truncation_error_aux with u0 u1 f A B; auto.
replace  (Z_of_nat (S k)) with (Z_of_nat 1 +Z_of_nat k)%Z.
destruct H0;[left|right].
apply Zlt_le_trans with (1:=H0);apply Zeq_le; ring.
apply Zle_lt_trans with (2:=H0);apply Zeq_le; ring.
rewrite (inj_S k); simpl (Z_of_nat 1); unfold Zsucc; ring.
unfold sum_f in *; simpl in *.
rewrite Rmult_plus_distr_l.
apply Rplus_le_compat.
replace (n-0)%nat with n in IHn by auto with arith.
apply IHn.
replace (S (n + 1)) with ((n+1-0)+1)%nat.
apply minorationenergy2 with (fun _ => 0) 
   (fun j0 : Z => convergence_error usol unh dX j0 1 / snd dX)
    (fun (j0 : Z) (m : nat) =>
          truncation_error c usol u0h u1h fh dX j0 (S m))
    (up (c * snd dX / fst dX))
    (Int_part (A/fst dX)-(up (c * snd dX / fst dX)))%Z (Int_part (B/fst dX)+(up (c * snd dX / fst dX)))%Z; trivial.
unfold is_discrete_solution; intros.
apply convergence_error_is_discrete_sol; auto with real.
apply up_gt_zero.
repeat apply Rmult_lt_0_compat; auto with real.
intros; rewrite FS_convergence_error_aux; trivial.
unfold Rdiv; ring.
replace (Z_of_nat 1 * up (c * snd dX / fst dX))%Z with (up (c * snd dX / fst dX)); try exact H0.
simpl (Z_of_nat 1); ring.
intros; apply FS_truncation_error_aux with u0 u1 f A B; auto.
replace  (Z_of_nat (S k)) with (Z_of_nat 1 +Z_of_nat k)%Z.
destruct H0;[left|right].
apply Zlt_le_trans with (1:=H0);apply Zeq_le; ring.
apply Zle_lt_trans with (2:=H0);apply Zeq_le; ring.
rewrite (inj_S k); simpl (Z_of_nat 1); unfold Zsucc; ring.
omega.
apply Rle_ge; apply Rmult_le_pos.
apply Rmult_le_pos.
unfold Rdiv; apply Rmult_le_pos.
apply sqrt_pos.
left; apply Rinv_0_lt_compat.
apply sqrt_lt_R0.
replace (2*eps-Rsqr eps) with (eps*(2-eps)) by (unfold Rsqr; ring).
apply Rmult_lt_0_compat.
apply eps_lt.
apply Rplus_lt_reg_r with eps.
ring_simplify.
apply Rlt_trans with 1; try apply eps_lt; auto with real.
now left.
unfold sum_f; simpl.
apply cond_pos_sum.
intros; apply sqrt_pos.
apply OuP_eq_trans_l with (scalar_multiplyAA_fun _ _
  (sqrt 2 / sqrt (2*eps-Rsqr eps))
  (fun (t : {t : R | Under T t}) (dX : R * R) =>
    snd dX *
    sum_f 1 (kt (projT1 t) (snd dX))
     (fun n : nat =>
      sqrt (energy c (convergence_error usol unh dX) dX (n - 1))))).
intros.
unfold scalar_multiplyAA_fun; ring.
apply OuP_multiplyAA.
apply OuP_sum_f.
apply T_pos.
intros dH (_,(T,_)); exact T.
apply energie; trivial.
Qed.

End Conv.

