(** The wave equation: its definition, its discretization *)

Require Export Differential.
Require Export R_n.
Require Export ZArith.
Open Local Scope nat_scope.
Open Local Scope R_scope.

Variable T:R.
Hypothesis T_pos: 0 < T.

Section Wave.
(** 1D wave equation *)

Variable c:R.
Variable u0 u1: R->R.
Variable f : R*R -> R.
Variable usol : R * R -> R.

Variable u0h u1h: Z-> R.
Variable fh: Z -> nat -> R.
Variable unh : R * R -> Z -> nat -> R.
Variable eps eta:R.

Definition L (u : R * R -> R) (X : R * R) : R :=
  (partial_derive_secondvar_n 2%nat u) X -
  c ^ 2 * (partial_derive_firstvar_n 2%nat u) X.

Definition L1  (u: R*R -> R) (X:R*R) : R :=
   (partial_derive_secondvar u) X.

Definition L0  (u: R*R -> R) (X:R*R) : R :=
  u X.

Definition is_solution (u: R*R -> R)   :=
  (forall X:R*R, L u X=f X) /\ 
  (forall x:R, L1 u (x,0) = u1 x) /\
  (forall x:R, L0 u (x,0) = u0 x).


Hypothesis usol_is_sol: is_solution usol.

(** Discretization *)

Definition discretization (u : R * R -> R) (dX : R * R) : Z -> nat-> R :=
  fun j k => u (Xjk dX j k).

Definition ubar (dX : R * R) (j:Z) (k : nat) : R :=
  discretization usol dX j k. 

(** Numerical scheme *)

Definition Ah (unh : Z -> R) (dx : R) (j : Z) : R :=
  - (c *c) * (unh (j + 1)%Z - 2 * unh j + unh (j - 1)%Z) / (dx *dx).

Definition Lh (uh: Z -> nat -> R) (dX : R * R) (j:Z) (n : nat) : R :=
  match n with
   O => uh j 0%nat
 | 1 => (uh j 1%nat - uh j 0%nat)/ (snd dX) + (snd dX) / 2 * (Ah (fun  j' => uh j' 0%nat) (fst dX) j)
 | S (S  m) =>  (uh j (m + 2)%nat - 2 * uh j (m+1)%nat + uh j m) / (snd dX *snd dX) + 
                      (Ah (fun j' => uh j' (m+1)%nat) (fst dX) j)
  end.

Definition LL1h (v: R*R->R) (X dX:R*R) : R:=
   (v (fst X , snd X+snd dX) - v X)/ (snd dX) - partial_derive_secondvar v X -
   (snd dX) / 2 * (c*c) *(v (fst X +fst dX, snd X) - 2*v X + 
   v (fst X -fst dX, snd X))/(fst dX*fst dX).

(** Energy and errors *)

Definition energy (uh : Z -> nat -> R) (dX : R * R) (n : nat) :=
  1/2 *
    infinite_squared_norm_dx (fst dX)
      (scalar_multiply_n (1 / snd dX)
        (subtract_n
          (fun j => uh j (n + 1)%nat)
          (fun j => uh j n))) +
  1/2 *
    infinite_dot_product_dx (fst dX)
      (fun j => (Ah (fun j' => uh j' n) (fst dX) j)) 
      (fun j => uh j (n + 1)%nat).



Definition scheme (uh: R*R -> Z -> nat -> R) (dX : R * R) (j:Z) (n : nat) : R :=
   match n with
   O => Lh (uh dX) dX j 0 - u0h j 
 | 1 => Lh (uh dX) dX j 1 - u1h j
 | S (S m) => Lh (uh dX) dX j (S (S m)) - fh j (S m)
  end.

Definition is_discrete_solution (dX:R*R) := forall  j n, 
   scheme unh dX j n = 0.

Definition truncation_error := (scheme ubar).


Definition convergence_error (dX : R * R) (j:Z) (n : nat) : R :=
  (ubar dX j n) - (unh dX j n).


(** A few properties, mainly of Ah *)

Lemma FS_Ah: forall u dx, FS u -> FS (Ah u dx).
intros u dx H; unfold Ah.
apply FS_eq with (scalar_multiply_n (-c*c/(dx*dx)) (add_n (fun j => u (j+1)%Z)
  (add_n (scalar_multiply_n (-2) u) (fun j => u (j-1)%Z)))).
auto with FS.
intros; unfold scalar_multiply_n, add_n, Rdiv; ring.
Qed.

Lemma FS_Ah_fun: forall u dx, FS u -> FS (fun j => Ah u dx j).
apply FS_Ah.
Qed.

Hint Resolve FS_Ah FS_Ah_fun: FS.

Lemma Ah_is_linear: forall uh1 uh2 dx j,
  Ah (fun j0 => uh1 j0 - uh2 j0) dx j = Ah uh1 dx j - Ah uh2 dx j.
intros; unfold Ah; unfold Rdiv; ring.
Qed.


Lemma Lh_is_linear: forall uh1 uh2 dX j n,
  Lh (fun j0 n0 => uh1 j0 n0 - uh2 j0 n0) dX j n = Lh uh1 dX j n - Lh uh2 dX j n.
intros; unfold Lh; case n; auto; intros; case n0.
rewrite Ah_is_linear; unfold Rdiv; ring.
intros; rewrite Ah_is_linear; unfold Rdiv; ring.
Qed.

Lemma Green_formula: forall dx u v,
  FS u -> FS v ->
 infinite_dot_product_dx dx (fun n => u (n+1)%Z -2* u n + u (n-1)%Z) v
  = - infinite_dot_product_dx dx (fun n => u (n+1)%Z - u n)
                                                     (fun n => v (n+1)%Z - v n).
intros.
rewrite infinite_dot_product_dx_eq_l with  
  (v:= (add_n (fun j => u (j + 1)%Z ) 
        (add_n (scalar_multiply_n (-2) (fun j => u j))
              (fun j => u (j - 1)%Z )))); auto with FS.
2: unfold scalar_multiply_n, add_n; intros; ring.
rewrite infinite_dot_product_dx_eq_l with 
  (u:=(fun n : Z => u (n + 1)%Z - u n)) 
  (v:=subtract_n (fun n => u (n+1)%Z) u); FS.
rewrite infinite_dot_product_dx_eq_r with 
  (u:=(fun n : Z => v (n + 1)%Z - v n)) 
  (v:=subtract_n (fun n => v (n+1)%Z) v); FS.
repeat rewrite infinite_dot_product_dx_add_l; FS.
rewrite infinite_dot_product_dx_sub_l; FS.
repeat rewrite infinite_dot_product_dx_sub_r; FS.
rewrite infinite_dot_product_dx_scal_mult_l; FS.
rewrite <- infinite_dot_product_dx_shift_indices; FS.
rewrite infinite_dot_product_dx_shift_indices with 
  (f1:=(fun j : Z => u (j - 1)%Z)); FS.
rewrite infinite_dot_product_dx_eq_l with 
  (v:=u) (u:=(fun j : Z => u (j + 1 - 1)%Z)); FS.
2: apply FS_shift1 with (l:=1%Z) (f:= fun j => u (j-1)%Z); FS.
2: intros; apply f_equal; auto with zarith.
rewrite infinite_dot_product_dx_eq_l with 
  (v:=u) (u:=(fun j : Z => u j)); FS.
ring.
Qed.

Lemma Ah_is_symmetrical :
  forall dx u v, FS u -> FS v ->
    infinite_dot_product_dx dx (Ah u dx) v
   = infinite_dot_product_dx dx u (Ah v dx).
intros; unfold Ah.
rewrite infinite_dot_product_dx_eq_l with  (v:=(scalar_multiply_n (-c*c/(dx*dx))
   (fun n => u (n + 1)%Z - 2 * u n + u (n - 1)%Z))); FS.
2: apply FS_eq with (Ah u dx); trivial.
2: now apply FS_Ah.
2: unfold scalar_multiply_n, Rdiv; intros; ring.
rewrite infinite_dot_product_dx_eq_r with  (v:=(scalar_multiply_n (-c*c/(dx*dx))
   (fun n => v (n + 1)%Z - 2 * v n + v (n - 1)%Z))) (w:=u); FS.
2: apply FS_eq with (Ah v dx); trivial.
2: now apply FS_Ah.
2: unfold scalar_multiply_n, Rdiv; intros; ring.
rewrite infinite_dot_product_dx_sym with (u:=u); FS.
repeat rewrite infinite_dot_product_dx_scal_mult_l; FS.
apply f_equal.
repeat rewrite Green_formula; FS.
rewrite infinite_dot_product_dx_sym; FS; auto with real.
Qed.

Lemma Ah_eq: forall uh1 uh2, (forall j, uh1  j=uh2  j) -> 
  forall dX j, Ah uh1 dX j= Ah uh2 dX j.
intros;unfold Ah.
now repeat rewrite H.
Save.


Lemma Ah_cont: forall dx vh,
  0 < dx -> FS vh ->
  infinite_dot_product_dx dx  (Ah vh dx) vh 
      <= 4*c*c/(dx*dx)*infinite_squared_norm_dx dx vh. 
intros; unfold Ah.
rewrite infinite_dot_product_dx_eq_l with  (v:=(scalar_multiply_n ((c*c/(dx*dx))*(-1))
   (fun n => vh (n + 1)%Z - 2 * vh n + vh (n - 1)%Z))); FS.
2: apply FS_eq with (Ah vh dx); FS.
2: unfold scalar_multiply_n, Rdiv; intros; ring.
rewrite infinite_dot_product_dx_scal_mult_l; FS.
rewrite Rmult_assoc.
apply Rle_trans with
  ((c * c / (dx * dx))*(4 * infinite_squared_norm_dx dx vh));[idtac|right; unfold Rdiv; ring].
apply Rmult_le_compat_l.
apply Rmult_le_pos.
fold (Rsqr c); auto with real.
left; apply Rinv_0_lt_compat; auto with real.
fold (Rsqr dx); auto with real.
rewrite Green_formula; FS.
ring_simplify.
fold (infinite_squared_norm_dx dx  (fun n : Z => vh (n + 1)%Z - vh n)).
apply Rle_trans with (2 * 
  (infinite_squared_norm_dx dx (fun n => vh (n+1)%Z) + infinite_squared_norm_dx dx vh)).
apply infinite_squared_norm_parall_sub; FS.
auto with real.
replace (infinite_squared_norm_dx dx (fun n : Z => vh (n + 1)%Z)) with
    (infinite_squared_norm_dx dx vh).
right; ring.
unfold infinite_squared_norm_dx.
now rewrite <- infinite_dot_product_dx_shift_indices.
Qed.

Lemma Ah_pos: forall dx vh,
  0 < dx -> FS vh ->
  0 <= infinite_dot_product_dx dx  (Ah vh dx) vh.
intros; unfold Ah.
rewrite infinite_dot_product_dx_eq_l with  (v:=(scalar_multiply_n (-c*c/(dx*dx))
   (fun n => vh (n + 1)%Z - 2 * vh n + vh (n - 1)%Z))); FS.
2: apply FS_eq with (Ah vh dx); trivial.
2: now apply FS_Ah.
2: unfold scalar_multiply_n, Rdiv; intros; ring.
rewrite infinite_dot_product_dx_scal_mult_l; FS.
rewrite Green_formula; FS.
rewrite Ropp_mult_distr_r_reverse.
unfold Rdiv; repeat rewrite Ropp_mult_distr_l_reverse.
rewrite Ropp_involutive.
apply Rmult_le_pos.
apply Rmult_le_pos.
fold (Rsqr c); auto with real.
left; apply Rinv_0_lt_compat.
apply Rmult_lt_0_compat; auto with real.
apply infinite_squared_norm_dx_pos; FS.
left; trivial.
Qed.


(** Scheme properties *)

Hypothesis eps_lt: 0 < eps < 1.
Hypothesis eta_pos: 0 < eta.

Definition CFL (dX:R*R)  := c * (snd dX) / (fst dX).

Definition CFL_condition_r (dX:R*R) := CFL dX <=  1 - eps.

Definition CFL_condition_l (dX:R*R) := eta <= CFL dX.

Notation OupsCFL_l := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX /\ CFL_condition_l dX)).
Notation OupsCFL := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX /\ CFL_condition_r dX /\ CFL_condition_l dX)).

Definition scheme_is_consistent_of_order (p q : nat) :=    
    OupsCFL_l (fun (t: {t| Under T t}) dX => infinite_squared_norm_dx (fst dX)
                          (fun j => truncation_error dX j  (kt (projT1 t) (snd dX))))
      (fun dX => Rsqr ((fst dX) ^ p + (snd dX) ^ q)).

Definition scheme_is_convergent_of_order (p q : nat) :=    
    OupsCFL (fun (t: {t| Under T t}) dX => infinite_norm_dx (fst dX)
                          (fun j => convergence_error dX j  (kt (projT1 t) (snd dX))))
      (fun dX =>  (fst dX) ^ p + (snd dX) ^ q).

End Wave.

(** The discretization has a finite support *)

Section zeroness.

Variable c:R.
Variable u0h u1h: Z-> R.
Variable fh: Z -> nat -> R.
Variable unh : R * R -> Z -> nat -> R.
Variable dX:R*R.
Variable l:Z.
Variable j1 j2:Z.

Hypothesis snd_dX_neq_0: snd dX <> 0.

Hypothesis unh_is_discrete_sol: is_discrete_solution c u0h u1h fh unh dX.

Hypothesis l_pos: (0 < l)%Z.

Hypothesis u0h_zero: forall j, (j < j1)%Z  \/ ( j2 < j)%Z 
   -> u0h j = 0.
Hypothesis u1h_zero: forall j, (j < j1)%Z  \/ ( j2 < j)%Z 
   -> u1h j = 0.
Hypothesis fh_zero: forall j k, (j < j1-Z_of_nat k*l)%Z  \/ ( j2+Z_of_nat k*l < j)%Z 
   -> fh j k= 0.


Lemma zeroness_unh_h: forall (n:nat) j,
  (j < j1 - Z_of_nat n*l)%Z  \/ (j2 + Z_of_nat n*l < j)%Z 
    -> unh dX j n = 0.
intros n.
apply lt_wf_ind with (P:= fun n => forall j, (j < j1 - Z_of_nat n*l)%Z 
  \/ (j2 + Z_of_nat n*l < j)%Z -> unh dX j n = 0).
clear n; intros n; case n.
(* n = 0 *)
intros R; clear R; simpl; intros.
elim (unh_is_discrete_sol j 0%nat).
unfold scheme, Lh.
rewrite u0h_zero; auto with real zarith.
case H;[left|right]; auto with zarith.
clear n; intros n; case n.
(* n = 1 *)
simpl (Z_of_nat 1); intros.
generalize (unh_is_discrete_sol j 1%nat).
unfold scheme, Lh, Ah.
repeat rewrite (H 0%nat); try (simpl; omega).
rewrite u1h_zero; try omega.
replace ((unh dX j 1 - 0) / snd dX +
snd dX / 2 * (- (c * c) * (0 - 2 * 0 + 0) / (fst dX * fst dX)) - 0) with
  (unh dX j 1 / snd dX).
intros.
apply Rmult_eq_reg_l with (/ snd dX).
ring_simplify.
rewrite <- H1; unfold Rdiv; ring.
auto with real.
unfold Rdiv; ring.
(* n > 1 *)
clear n; intros.
repeat rewrite inj_S in H0; unfold Zsucc in H0.
generalize (unh_is_discrete_sol j (S (S n))).
unfold scheme, Lh, Ah.
rewrite fh_zero.
repeat rewrite (H (n+1)%nat); try omega.
rewrite (H n); try omega.
replace ((unh dX j (n + 2) - 2 * 0 + 0) / (snd dX * snd dX) +
- (c * c) * (0 - 2 * 0 + 0) / (fst dX * fst dX) - 0) with  
  (unh dX j (n + 2) / (snd dX * snd dX)).
intros.
apply Rmult_eq_reg_l with (/ (snd dX * snd dX)).
replace (S (S n)) with (n+2)%nat by omega.
ring_simplify.
rewrite <- H1; unfold Rdiv; ring.
auto with real.
unfold Rdiv; ring.
assert (Z_of_nat n * l <= (Z_of_nat n + 1 + 1) * l)%Z; 
  [apply Zmult_le_compat_r|idtac]; omega.
rewrite inj_plus; simpl.
assert ((Z_of_nat n+1) * l < (Z_of_nat n + 1+1) * l)%Z; 
  [apply Zmult_lt_compat_r|idtac]; try omega.
rewrite inj_plus; simpl.
assert ((Z_of_nat n+1) * l < (Z_of_nat n + 1 + 1) * l)%Z; 
  [apply Zmult_lt_compat_r|idtac]; omega.
rewrite inj_plus; simpl.
assert ((Z_of_nat n+1) * l < (Z_of_nat n + 1 + 1) * l)%Z; 
  [apply Zmult_lt_compat_r|idtac]; omega.
rewrite inj_S; simpl.
assert (Zsucc(Z_of_nat n) * l < (Z_of_nat n + 1 + 1) * l)%Z; 
  [apply Zmult_lt_compat_r|idtac]; omega.
Qed.


End zeroness.

Section zeroness2.

Variable c:R.
Variable u0 u1: R->R.
Variable f : R*R -> R.
Variable u0h u1h: Z-> R.
Variable fh: Z -> nat -> R.
Variable unh : R * R -> Z -> nat -> R.
Variable dX:R*R.
Variable A B:R.

Hypothesis c_pos: 0 < c.
Hypothesis fst_dX_pos: 0 < fst dX.
Hypothesis snd_dX_pos: 0 < snd dX.

Hypothesis unh_is_discrete_sol: is_discrete_solution c u0h u1h fh unh dX.


Hypothesis u0_zero: forall x, (x < A)  \/ (B < x) 
   -> u0 x = 0.
Hypothesis u1_zero: forall x, (x < A) \/ (B < x) 
   -> u1 x = 0.
Hypothesis f_zero: forall x t, (x < A-c*t)  \/ (B +c*t< x) 
   -> f (x,t) = 0.

Hypothesis u0h_is_u0: forall dX j, u0h j = u0 (fst (Xjk dX j 0)).
Hypothesis u1h_is_u1: forall dX j, u1h j = u1 (fst (Xjk dX j 1)).
Hypothesis fh_is_f: forall dX j k, fh j k = f (Xjk dX j k).

Lemma zeroness_unh_aux1:
  (forall j l, (j < Int_part (A / fst dX))%Z \/ (Int_part (B / fst dX) < j)%Z
   -> fst (Xjk dX j l) < A \/ B < fst (Xjk dX j l)).
clear -fst_dX_pos snd_dX_pos; intros; case H; clear H; intros.
left; unfold Xjk; simpl.
apply Rmult_lt_reg_l with (/(fst dX)).
auto with real.
replace (/ fst dX * (IZR j * fst dX)) with (IZR j) by (field; auto with real).
apply Rlt_le_trans with (IZR ((Int_part (A / fst dX)))).
apply IZR_lt; auto with zarith.
apply Rle_trans with ((A / fst dX)).
eapply base_Int_part.
unfold Rdiv; right; ring.
right; unfold Xjk; simpl.
apply Rmult_lt_reg_l with (/(fst dX)).
auto with real.
replace (/ fst dX * (IZR j * fst dX)) with (IZR j) by (field; auto with real).
apply Rlt_le_trans with (IZR ((Int_part (B / fst dX))+1)).
2:apply IZR_le; auto with zarith.
apply Rle_lt_trans with ((B / fst dX)).
unfold Rdiv; right; ring.
rewrite plus_IZR; simpl.
apply Rplus_lt_reg_r with (-1-B / fst dX); ring_simplify.
rewrite Rplus_comm.
eapply base_Int_part.
Qed.


Lemma zeroness_unh_aux2: forall j k,
 (j < Int_part (A / fst dX) - Z_of_nat k*(up (c*snd dX/ fst dX)))%Z \/
    (Int_part (B / fst dX) + Z_of_nat k*(up (c*snd dX/ fst dX)) < j)%Z ->
 fst (Xjk dX j k) < A - c * snd (Xjk dX j k) \/
  B + c * snd (Xjk dX j k) < fst (Xjk dX j k).
intros; simpl.
case H; intros.
left.
apply Rmult_lt_reg_l with (/fst dX).
auto with real.
replace (/ fst dX * (IZR j * fst dX)) with (IZR j).
2: field; auto with real.
apply Rlt_le_trans with (IZR ( Int_part (A/fst dX) - Z_of_nat k* (up (c*snd dX / fst dX)))).
now apply IZR_lt.
unfold Zminus; rewrite plus_IZR.
unfold Rminus; rewrite Rmult_plus_distr_l.
apply Rplus_le_compat.
rewrite Rmult_comm.
eapply base_Int_part.
rewrite Ropp_Ropp_IZR.
rewrite Ropp_mult_distr_r_reverse.
apply Ropp_le_contravar.
rewrite mult_IZR.
rewrite <- INR_IZR_INZ.
apply Rle_trans with (INR k*(c*snd dX/fst dX)).
right; unfold Rdiv; ring.
apply Rmult_le_compat_l; auto with real.
left; eapply archimed.
right.
apply Rmult_lt_reg_l with (/fst dX).
auto with real.
replace (/ fst dX * (IZR j * fst dX)) with (IZR j).
2: field; auto with real.
apply Rlt_le_trans with (IZR ( (Int_part (B/fst dX)+1) + Z_of_nat k*(up (c * snd dX / fst dX)))).
2: apply IZR_le; auto with zarith.
rewrite plus_IZR.
rewrite Rmult_plus_distr_l.
apply Rlt_le_trans with 
  (IZR (Int_part (B/fst dX) + 1) + / fst dX * (c * (INR k * snd dX))).
apply Rplus_lt_compat_r.
rewrite Rmult_comm.
rewrite plus_IZR; simpl.
apply Rplus_lt_reg_r with (-1-B*/fst dX); ring_simplify.
rewrite Rplus_comm.
apply Rlt_le_trans with 
  (IZR (Int_part (B / fst dX)) - B / fst dX).
eapply base_Int_part.
right; unfold Rdiv;ring.
apply Rplus_le_compat_l.
rewrite mult_IZR.
rewrite <- INR_IZR_INZ.
apply Rle_trans with (INR k*(c*snd dX/fst dX)).
right; unfold Rdiv; ring.
apply Rmult_le_compat_l; auto with real.
left; eapply archimed.
Qed.



Lemma zeroness_unh: forall (n:nat) j,
  (j < Int_part (A/fst dX) - Z_of_nat n *(up (c*snd dX/ fst dX)))%Z  
     \/ ( Int_part (B/fst dX) + Z_of_nat n *(up (c*snd dX/ fst dX))< j)%Z 
  -> unh dX j n = 0.
intros.
(* . *)
apply zeroness_unh_h with c u0h u1h fh (up (c*snd dX/ fst dX))%Z
(Int_part (A/fst dX)) (Int_part (B/fst dX)); trivial.
auto with real.
assert (0 <= Int_part (c * snd dX / fst dX))%Z.
apply Int_part_pos.
repeat apply Rmult_le_pos; left; auto with real.
apply up_gt_zero.
unfold Rdiv; repeat apply Rmult_lt_0_compat; auto with real.
clear H j;intros.
rewrite (u0h_is_u0 dX).
apply u0_zero.
now apply zeroness_unh_aux1.
clear H j;intros.
rewrite (u1h_is_u1 dX).
apply u1_zero.
now apply zeroness_unh_aux1.
clear H j;intros.
rewrite (fh_is_f dX).
replace (Xjk dX j k) with ((fst ((Xjk dX j k))), snd ((Xjk dX j k))).
apply f_zero.
now apply zeroness_unh_aux2.
reflexivity.
Qed.


End zeroness2.

(** The exact solution has finite support *)

Section zeronessusol.

Variable c:R.
Variable u0 u1: R->R.
Variable f : R*R -> R.
Variable usol : R * R -> R.
Variable A B : R.

Hypothesis usol_is_sol: is_solution c u0 u1 f usol.

Hypothesis u0_zero: forall x, (x < A)  \/ (B < x)  -> u0 x = 0.
Hypothesis u1_zero: forall x, (x < A)  \/ (B < x)  -> u1 x = 0.
Hypothesis f_zero: forall x t, (x < A-c*t)  \/ (B +c*t< x) 
   -> f (x,t) = 0.

Axiom zeroness_usol: 
  is_solution c u0 u1 f usol ->
  (forall x, (x < A)  \/ (B < x)  -> u0 x = 0) ->
  (forall x, (x < A)  \/ (B < x)  -> u1 x = 0) ->
  (forall x t, (x < A-c*t)  \/ (B +c*t< x) -> f (x,t) = 0) ->
   forall X,
     (fst X < A-c*snd X)  \/ (B +c*snd X< fst X) -> usol X=0.

End zeronessusol.




